#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import datetime
import logging
import copy
import pytest

from hydra_base import HydraError
from hydra_base.lib.objects import JSONObject

log = logging.getLogger(__name__)

class TestProjectInheritance:
    """
        Test for working with project inheritance in Hydra
    """

    def test_add_child_project(self, client):
        """
            Test adding a child project to a project
        """
        parentproject = JSONObject({})
        parentproject.name = 'Test Parent %s'%(datetime.datetime.now())

        new_project_i = client.add_project(parentproject)


        childproject = JSONObject({})
        childproject.name = 'Test Child %s'%(datetime.datetime.now())
        childproject.parent_id = new_project_i.id

        child_project_j = client.add_project(childproject)

        parentproject_j = client.get_project(new_project_i.id)

        assert len(parentproject_j.projects) == 1
        assert parentproject_j.projects[0].id == child_project_j.id

    def test_move_project_with_update(self, client):
        """
            Test adding a child project to a project
        """
        #Add another parent that the child will be moved to
        targetproject = JSONObject({})
        targetproject.name = 'Target Parent %s'%(datetime.datetime.now())
        target_project_i = client.add_project(targetproject)

        childproject = JSONObject({})
        childproject.name = 'Test Child %s'%(datetime.datetime.now())
        child_project_j = client.add_project(childproject)

        #make sure the child isn't in the parent
        sourceproject_j = client.get_project(target_project_i.id)
        assert len(sourceproject_j.projects) == 0


        #Update the child to say it's now in the parent.
        child_project_j.parent_id = target_project_i.id
        client.update_project(child_project_j)

        ##make sure the child is in the target
        sourceproject_j = client.get_project(target_project_i.id)
        assert len(sourceproject_j.projects) == 1
        assert sourceproject_j.projects[0].id == child_project_j.id

    def test_move_project(self, client):
        """
            Test adding a child project to a project
        """
        #Add a parent
        sourceproject = JSONObject({})
        sourceproject.name = 'Source Parent %s'%(datetime.datetime.now())
        source_project_i = client.add_project(sourceproject)

        #Add another parent that the child will be moved to
        targetproject = JSONObject({})
        targetproject.name = 'Target Parent %s'%(datetime.datetime.now())
        target_project_i = client.add_project(targetproject)

        childproject = JSONObject({})
        childproject.name = 'Test Child %s'%(datetime.datetime.now())
        childproject.parent_id = source_project_i.id
        child_project_j = client.add_project(childproject)

        #Make sure the child is added coorectly
        parentproject_j = client.get_project(source_project_i.id)
        assert len(parentproject_j.projects) == 1
        assert parentproject_j.projects[0].id == child_project_j.id

        #Move the project to another project
        client.move_project(child_project_j.id, target_project_i.id)

        ##make sure the child is no longer in the original parent
        sourceproject_j = client.get_project(source_project_i.id)
        assert len(sourceproject_j.projects) == 0

        ##check the child is in the destination project as requested
        targetproject_j = client.get_project(target_project_i.id)
        assert len(targetproject_j.projects) == 1
        assert targetproject_j.projects[0].id == child_project_j.id


    def test_share_project(self, client, projectmaker):
        """
            Test sharing a sub-project. This should result in the sharee having
            access to the full tree of projects until the shared projects, but not
            the projects which are siblings of any projects in that path through the tree
            For example, upon sharing p4, the sharee should not have access to p3
                                 p1
                                /  \
                               p2   p3
                              /
                             p4
        """

        proj_user = client.user_id
        proj1 = projectmaker.create(share=False)
        proj2 = projectmaker.create(share=False, parent_id=proj1.id)
        proj3 = projectmaker.create(share=False, parent_id=proj1.id)
        proj4 = projectmaker.create(share=False, parent_id=proj2.id)

        client.user_id = pytest.user_c.id
        with pytest.raises(HydraError):
            client.get_project(proj4.id)

        #Now, as the main user, share P4 with user C
        client.user_id = proj_user
        client.share_project(proj4.id, ['UserC'], False, False)

        #Now as the sharee, try to get project c
        client.user_id = pytest.user_c.id
        userc_p4 = client.get_project(proj4.id)
        userc_p1 = client.get_project(proj1.id)
        userc_p2 = client.get_project(proj2.id)

        #Check user can't see p3 which is in p1, but doesn't have correct permissions
        p1_children = [p.id for p in userc_p1.projects]
        assert p1_children == [proj2.id]

        #Check user can't see p3 which is in p1, but doesn't have correct permissions
        p2_children = [p.id for p in userc_p2.projects]
        assert p2_children == [proj4.id]

        with pytest.raises(HydraError):
            client.get_project(proj3.id)


        client.user_id = proj_user
        client.unshare_project(proj4.id, ['UserC'])

        client.user_id = pytest.user_c.id
        with pytest.raises(HydraError):
            client.get_project(proj4.id)
        client.user_id = proj_user

    def test_share_network(self, client, projectmaker, networkmaker):
        """
            Test sharing a network contained in a sub-project. This should result in the sharee having
            access to the full tree of projects until the shared projects, but not
            the projects which are siblings of any projects in that path through the tree
            For example, upon sharing p4, the sharee should not have access to p3
                                 p1
                                /  \
                               p2   p3
                              /
                             p4
                            /
                           n1
        """
        client.user_id = 1 # force current user to be 1 to avoid potential inconsistencies
        proj_user = client.user_id
        proj1 = projectmaker.create(share=False)
        proj2 = projectmaker.create(share=False, parent_id=proj1.id)
        proj3 = projectmaker.create(share=False, parent_id=proj1.id)
        proj4 = projectmaker.create(share=False, parent_id=proj2.id)

        net1 = networkmaker.create(project_id=proj4.id)

        client.user_id = pytest.user_c.id
        with pytest.raises(HydraError):
            client.get_project(proj4.id)

        with pytest.raises(HydraError):
            client.get_network(net1.id)

        #Now, as the main user, share P4 with user C
        client.user_id = proj_user
        client.share_network(net1.id, ['UserC'], False, False)

        #Now as the sharee, try to get project c
        client.user_id = pytest.user_c.id
        userc_n1 = client.get_network(net1.id)
        userc_p4 = client.get_project(proj4.id)
        userc_p1 = client.get_project(proj1.id)
        userc_p2 = client.get_project(proj2.id)

        assert len(userc_p4.networks) == 1
        #Check user can't see p3 which is in p1, but doesn't have correct permissions
        p1_children = [p.id for p in userc_p1.projects]
        assert p1_children == [proj2.id]

        #Check user can't see p3 which is in p1, but doesn't have correct permissions
        p2_children = [p.id for p in userc_p2.projects]
        assert p2_children == [proj4.id]

        with pytest.raises(HydraError):
            client.get_project(proj3.id)

        client.user_id = proj_user
        client.unshare_network(net1.id, ['UserC'])

        client.user_id = pytest.user_c.id
        with pytest.raises(HydraError):
            client.get_network(net1.id)

        #user no longer has access to P4 because there's no shared network in there
        with pytest.raises(HydraError):
            client.get_project(proj4.id)
        client.user_id = proj_user

    def test_access_to_shared_sub_project(self, client, projectmaker):
        """
            Test sharing a sub-project. This should result in the sharee having
            access to the full tree of projects until the shared projects, but not
            the projects which are siblings of any projects in that path through the tree
            For example, upon sharing p4, the sharee should not have access to p3
                                 p1
                                /  \
                               p2   p3
                              /
                             p4
            If userC wants to navigate to p4, they will also need read access on P1 and P2
            while not being able to see p3 in the list of sub-projects of P1
        """

        proj_user = client.user_id
        proj1 = projectmaker.create(share=False, name='Project1')
        proj2 = projectmaker.create(share=False, name='Project2', parent_id=proj1.id)
        proj3 = projectmaker.create(share=False, name='Project3', parent_id=proj1.id)
        proj4 = projectmaker.create(share=False, name='Project4', parent_id=proj2.id)

        client.user_id = pytest.user_c.id
        with pytest.raises(HydraError):
            client.get_project(proj4.id)

        #Now, as the main user, share P4 with user C
        client.user_id = proj_user
        client.share_project(proj4.id, ['UserC'], False, False)

        #Now as the sharee, try to get project c
        client.user_id = pytest.user_c.id
        userc_projects = client.get_projects(pytest.user_c.id)
        assert proj1.id in [p.id for p in userc_projects]

        #User C doesn't have explicit read access on proj1 or proj2, but should
        #be abe to navigate to proj1 and 2 so they cna access proj4
        userc_proj1 = client.get_project(project_id=proj1.id)

        #User C should only see project 2, not project 3, as they don't hace access
        #to that branch
        assert len(userc_proj1.projects) == 1

        assert proj2.id in [p.id for p in userc_proj1.projects]

        #User C can't see project 3
        with pytest.raises(HydraError):
            client.get_project(proj3.id)
        client.user_id = proj_user

    def test_access_to_shared_network_in_sub_project(self, client, projectmaker, networkmaker):
        """
            Test sharing a network contained in sub-project. This should result in the sharee having
            access to the full tree of projects until the shared projects, but not
            the projects which are siblings of any projects in that path through the tree
            For example, upon sharing p4, the sharee should not have access to p3
                                 p1
                                /  \
                               p2   p3
                              /
                             N1
            If userC wants to navigate to N1, they will also need read access on P1 and P2
            while not being able to see p3 in the list of sub-projects of P1
        """

        proj_user = client.user_id
        proj1 = projectmaker.create(share=False, name='Project11')
        proj2 = projectmaker.create(share=False, name='Project22', parent_id=proj1.id)
        proj3 = projectmaker.create(share=False, name='Project33', parent_id=proj1.id)

        net1 = networkmaker.create(project_id=proj2.id)

        client.user_id = pytest.user_c.id
        with pytest.raises(HydraError):
            client.get_network(net1.id)

        #Now, as the main user, share P4 with user C
        client.user_id = proj_user
        client.share_network(net1.id, ['UserC'], False, False)

        #Now as the sharee, try to get project c
        client.user_id = pytest.user_c.id
        userc_projects = client.get_projects(pytest.user_c.id)
        assert proj1.id in [p.id for p in userc_projects]

        #User C doesn't have explicit read access on proj1 or proj2, but should
        #be abe to navigate to proj1 and 2 so they cna access proj4
        userc_proj1 = client.get_project(project_id=proj1.id)

        #User C should only see project 2, not project 3, as they don't hace access
        #to that branch
        assert len(userc_proj1.projects) == 1

        assert proj2.id in [p.id for p in userc_proj1.projects]

        #User C can't see project 3
        with pytest.raises(HydraError):
            client.get_project(proj3.id)

        client.user_id = proj_user
        #Now unshare the network and check to make sure user C can no longer access the project tree.
        client.unshare_network(net1.id, ['UserC'])

        #Now as the sharee, try to get project c
        client.user_id = pytest.user_c.id
        userc_projects = client.get_projects(pytest.user_c.id)
        assert proj1.id not in [p.id for p in userc_projects]

        #User C no longer has access to project 1 as the network is no longer visible
        with pytest.raises(HydraError):
            client.get_project(project_id=proj1.id)

        #User C can't see project 3
        with pytest.raises(HydraError):
            client.get_project(project_id=proj3.id)

    def test_remove_project_parent(self, client, projectmaker, networkmaker):
        """
          Test two actions which should result in a project's parent_id
          being set to None:
            - Calling update_project() where the existing version has a not-None
              parent_id while the argument's parent_id has been set to None
            - Calling move_project() with a target parent project of None
        """
        orig_parent_proj = projectmaker.create()
        child_proj_update = projectmaker.create(parent_id=orig_parent_proj.id)
        child_proj_move = projectmaker.create(parent_id=orig_parent_proj.id)

        proj_i = client.get_project(child_proj_update.id)
        assert proj_i.parent_id == orig_parent_proj.id

        # Remove parent of child_proj
        proj_i.parent_id = None
        client.update_project(proj_i)

        # Verify retrieved project now has no parent
        proj_i = client.get_project(child_proj_update.id)
        assert proj_i.parent_id is None

        proj_i = client.get_project(child_proj_move.id)
        assert proj_i.parent_id == orig_parent_proj.id

        client.move_project(child_proj_move.id, None)

        # Verify moved project now has no parent
        proj_i = client.get_project(child_proj_move.id)
        assert proj_i.parent_id is None



"""template_inheritance

Revision ID: b04cb2e57cb0
Revises: 50bd5bac8701
Create Date: 2020-08-14 10:42:47.565737

"""
import logging
from alembic import op
import sqlalchemy as sa

log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'b04cb2e57cb0'
down_revision = '50bd5bac8701'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tTemplate', sa.Column('parent_id', sa.Integer(), sa.ForeignKey('tTemplate.id'), nullable=True))
        except Exception as e:
            log.critical(e)
        try:
            op.add_column('tTemplateType', sa.Column('parent_id', sa.Integer(), sa.ForeignKey('tTemplateType.id'), nullable=True))
        except Exception as e:
            log.critical(e)
        try:
            op.add_column('tTemplateType', sa.Column('status', sa.String(1), nullable=False, server_default=sa.text(u"'A'")))
        except Exception as e:
            log.critical(e)

            #change the primary key on tTypeAttr from (type_id, attr_id) to an auto-incremented (id)
        try:
            op.drop_constraint('tTypeAttr_ibfk_1', 'tTypeAttr', type_='foreignkey')
        except Exception as e:
            log.critical(e)
        try:
            op.drop_constraint('tTypeAttr_ibfk_2', 'tTypeAttr', type_='foreignkey')
        except Exception as e:
            log.critical(e)
        try:
            op.execute('alter table tTypeAttr drop primary key')
        except Exception as e:
            log.critical(e)
        try:
            op.execute('alter table tTypeAttr add column id int primary key auto_increment;')
        except Exception as e:
            log.critical(e)
        try:
            op.add_column('tTypeAttr', sa.Column('status', sa.String(1), nullable=False, server_default=sa.text(u"'A'")))
        except Exception as e:
            log.critical(e)

        try:
            op.add_column('tTypeAttr', sa.Column('parent_id', sa.Integer(), sa.ForeignKey('tTypeAttr.id'), nullable=True))
        except Exception as e:
            log.critical(e)

        try:
            op.create_foreign_key('tTypeAttr_ibfk_1', 'tTypeAttr', 'tAttr', ['attr_id'], ['id'])
        except Exception as e:
            log.critical(e)
        try:
            op.create_foreign_key('tTypeAttr_ibfk_2', 'tTypeAttr', 'tTemplateType', ['type_id'], ['id'])
        except Exception as e:
            log.critical(e)
        try:
            op.add_column('tResourceType', sa.Column('child_template_id', sa.Integer(), sa.ForeignKey('tTemplate.id'), nullable=True))
        except Exception as e:
            log.critical(e)



def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.drop_constraint('tTemplate_ibfk_1', 'tTypeAttr', type_='foreignkey')
            op.drop_column('tTemplate', 'parent_id')
            op.drop_constraint('tTemplateType_ibfk_2', 'tTypeAttr', type_='foreignkey')
            op.drop_column('tTemplateType', 'parent_id')


            op.drop_constraint(None, 'tTypeAttr', type_='primarykey')
            op.drop_constraint('tTypeAttr_ibfk_1', 'tTypeAttr', type_='foreignkey')
            op.drop_constraint('tTypeAttr_ibfk_2', 'tTypeAttr', type_='foreignkey')
            op.execute('alter table tTypeAttr add primary key (attr_id, type_id)')
            op.drop_column('tTypeAttr', 'parent_id')
            op.drop_column('tTypeAttr', 'id')
            op.drop_column('tTypeAttr', 'status')
            op.add_constraint('tTypeAttr_ibfk_1', 'tTypeAttr', type_='foreignkey')
            op.add_constraint('tTypeAttr_ibfk_2', 'tTypeAttr', type_='foreignkey')
            op.drop_column('tResourceType', 'child_template_id')
        except Exception as e:
            log.exception(e)



"""roles_perms

Revision ID: 21e8d704c020
Revises: d0c45fdc4fac
Create Date: 2018-04-17 17:22:12.357775

"""
from alembic import op
import sqlalchemy as sa


import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '21e8d704c020'
down_revision = 'd0c45fdc4fac'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tPerm

        try:
            op.alter_column('tPerm', 'perm_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tPerm', 'perm_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tPerm', 'perm_code', new_column_name='code', existing_type=sa.String(60), nullable=False)

            op.alter_column('tRole', 'role_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tRole', 'role_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
            op.alter_column('tRole', 'role_code', new_column_name='code', existing_type=sa.String(60), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tPerm
        try:
            op.drop_table('tPerm_new')
            op.drop_table('tRole_new')
        except:
            log.info("tPerm_new or tRole_new isn't there")

        try:
            # ## tPerm
            op.create_table(
                'tPerm_new',
                    sa.Column('id',sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('name',sa.String(60),  nullable=False),
                    sa.Column('code',sa.String(60),  nullable=False),
                    sa.Column('cr_date',sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
            )

            op.execute("insert into tPerm_new (id, name, code, cr_date) select perm_id, perm_name, perm_code, cr_date from tPerm")

            op.rename_table('tPerm','tPerm_old')
            op.rename_table('tPerm_new', 'tPerm')
            op.drop_table('tPerm_old')

            # ## tRole
            op.create_table(
                'tRole_new',
                    sa.Column('id',sa.Integer(), primary_key=True, nullable=False),
                    sa.Column('name',sa.String(60),  nullable=False),
                    sa.Column('code',sa.String(60),  nullable=False),
                    sa.Column('cr_date',sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
            )

            op.execute("insert into tRole_new (id, name, code, cr_date) select role_id, role_name, role_code, cr_date from tRole")

            op.rename_table('tRole','tRole_old')
            op.rename_table('tRole_new', 'tRole')
            op.drop_table('tRole_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import logging
import logging.config
from . import config
import os
import ctypes
import inspect

def init(level=None):
 #   if level is None:
 #       level = config.get('DEFAULT', 'log_level')

 #   if os.name == "nt":
 #       logging.addLevelName( logging.INFO, logging.getLevelName(logging.INFO))
 #       logging.addLevelName( logging.DEBUG, logging.getLevelName(logging.DEBUG))
 #       logging.addLevelName( logging.WARNING, logging.getLevelName(logging.WARNING))
 #       logging.addLevelName( logging.ERROR, logging.getLevelName(logging.ERROR))
 #       logging.addLevelName( logging.CRITICAL, logging.getLevelName(logging.CRITICAL))
 #       logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s', level=level)
 #       return

 #   if level is None:
 #       level = config.get('DEFAULT', 'log_level')


 #   logging.addLevelName( logging.INFO, "\033[0;m%s\033[0;m" % logging.getLevelName(logging.INFO))
 #   logging.addLevelName( logging.DEBUG, "\033[0;32m%s\033[0;32m" % logging.getLevelName(logging.DEBUG))
 #   logging.addLevelName( logging.WARNING, "\033[0;33m%s\033[0;33m" % logging.getLevelName(logging.WARNING))
 #   logging.addLevelName( logging.ERROR, "\033[0;31m%s\033[0;31m" % logging.getLevelName(logging.ERROR))
 #   logging.addLevelName( logging.CRITICAL, "\033[0;35m%s\033[0;35m" % logging.getLevelName(logging.CRITICAL))

 #   logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s\033[0m', level=level)
    #Log to a file with the same name as that calling 'import logging'
    #ex: server.py logs to server.log, ImportCSV.py logs to ImportCSV.log
    #All log files should be located in the same location.

    try:
        calling_file = inspect.stack()[-1][0].f_globals['__file__']
        calling_file = os.path.split(calling_file)[1]

        log_file = "%s.log" % calling_file.split('.')[0]
        log_base_path = config.get('logging_conf', 'log_file_dir', '.')

        if not os.path.isdir(log_base_path):
            os.makedirs(log_base_path)

        log_loc = os.path.expanduser(os.path.join(log_base_path, log_file))

    except:
        log_file = None
        log_loc  = None
        log_base_path = None

    use_default = False
    try:
        config_file = os.path.expanduser(config.get('logging_conf', 'log_config_path', '.'))
        #check the config file exists...
        if os.path.isfile(config_file) and log_base_path is not None:
            logging.config.fileConfig(config_file)
            logger = logging.getLogger()
            handler = logging.FileHandler(os.path.join(log_base_path, log_file),"a")
            handler.setLevel(logging.DEBUG)
            formatter = logging.Formatter("%(process)s %(asctime)s - %(levelname)s - %(message)s")
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            use_default=False
            logging.debug("Using logging config at %s", config_file)
        else:
            logging.debug("No logging config file found. Using default settings.")
            use_default = True
    except Exception as e:
        logging.exception(e)
        logging.critical("Error finding logging conf file: %s", e)
        use_default = True

    if use_default is True:
        logging_conf_dict = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'hydraFormatter': {
                'format': '%(process)s %(asctime)s - %(levelname)s - %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level':'DEBUG',
                'class':'hydra_base.hydra_logging.ColorizingStreamHandler',
                'formatter' : 'hydraFormatter',
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': 'INFO',
                'propagate': True
            },
        }
        }

        if log_loc is not None:
            logging_conf_dict['handlers']['file'] = {
                'level':'DEBUG',
                'class':'logging.FileHandler',
                'formatter' : 'hydraFormatter',
                'filename'  : log_loc,
                'mode'      : 'a',
            }

        logging.config.dictConfig(logging_conf_dict)

def shutdown():
	logging.shutdown()

class ColorizingStreamHandler(logging.StreamHandler):
#
#Thanks to Vinay Sajip
#(http://plumberjack.blogspot.co.uk/2010/12/colorizing-logging-output-in-terminals.html)
##

    @property
    def is_tty(self):
        isatty = getattr(self.stream, 'isatty', None)
        return isatty and isatty()

    def emit(self, record):
        try:
            message = self.format(record)
            stream = self.stream
            if not self.is_tty:
                stream.write("%s : %s"%(os.getpid(),message))
            else:
                self.output_colorized(message)
            stream.write(getattr(self, 'terminator', '\n'))
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def format(self, record):
        message = super(logging.StreamHandler, self).format(record)

        if self.is_tty:
            # Don't colorize any traceback
            parts = message.split('\n', 1)
            parts[0] = self.colorize(parts[0], record)
            message = '\n'.join(parts)
        return message

    # color names to indices
    color_map = {
        'black': 0,
        'red': 1,
        'green': 2,
        'yellow': 3,
        'blue': 4,
        'magenta': 5,
        'cyan': 6,
        'white': 7,
    }

    #levels to (background, foreground, bold/intense)
    if os.name == 'nt':
        level_map = {
            logging.DEBUG: (None, 'blue', True),
            logging.INFO: (None, 'white', False),
            logging.WARNING: (None, 'yellow', True),
            logging.ERROR: (None, 'red', True),
            logging.CRITICAL: ('red', 'white', True),
        }
    else:
        level_map = {
            logging.DEBUG: (None, 'cyan', False),
            logging.INFO: (None, 'white', False),
            logging.WARNING: (None, 'yellow', False),
            logging.ERROR: (None, 'red', False),
            logging.CRITICAL: ('red', 'white', True),
        }
    csi = '\x1b['
    reset = '\x1b[0m'

    def colorize(self, message, record):
        if record.levelno in self.level_map:
            bg, fg, bold = self.level_map[record.levelno]
            params = []
            if bg in self.color_map:
                params.append(str(self.color_map[bg] + 40))
            if fg in self.color_map:
                params.append(str(self.color_map[fg] + 30))
            if bold:
                params.append('1')
            if params:
                message = ''.join((self.csi, ';'.join(params),
                                   'm', message, self.reset))
        return message

    if os.name != 'nt':
        def output_colorized(self, message):
            self.stream.write(message)
    else:
        import re
        ansi_esc = re.compile(r'\x1b\[((\d+)(;(\d+))*)m')

        nt_color_map = {
            0: 0x00,    # black
            1: 0x04,    # red
            2: 0x02,    # green
            3: 0x06,    # yellow
            4: 0x01,    # blue
            5: 0x05,    # magenta
            6: 0x03,    # cyan
            7: 0x07,    # white
        }

        def output_colorized(self, message):
            parts = self.ansi_esc.split(message)
            write = self.stream.write
            h = None
            fd = getattr(self.stream, 'fileno', None)
            if fd is not None:
                fd = fd()
                if fd in (1, 2): # stdout or stderr
                    h = ctypes.windll.kernel32.GetStdHandle(-10 - fd)
            while parts:
                text = parts.pop(0)
                if text:
                    write(text)
                if len(parts) > 4:
                    params = parts[0]
                    parts = parts[4:]
                    if h is not None:
                        params = [int(p) for p in params.split(';')]
                        color = 0
                        for p in params:
                            if 40 <= p <= 47:
                                color |= self.nt_color_map[p - 40] << 4
                            elif 30 <= p <= 37:
                                color |= self.nt_color_map[p - 30]
                            elif p == 1:
                                color |= 0x08 # foreground intensity on
                            elif p == 0: # reset to default color
                                color = 0x07
                            else:
                                pass # error condition ignored
                        ctypes.windll.kernel32.SetConsoleTextAttribute(h, color)



# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2020 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import json
import logging
import re
from decimal import Decimal
import datetime

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import noload, joinedload

from hydra_base import db
from hydra_base.db.model import (Template, TemplateType, TypeAttr, Attr,
                                Network, Node, Link, ResourceGroup,
                                ResourceType, ResourceAttr, ResourceScenario, Scenario)

from hydra_base.db.model import Dataset as ModelDataset
from hydra_base.lib.objects import JSONObject, Dataset
from hydra_base.lib.data import add_dataset
from hydra_base.exceptions import HydraError, ResourceNotFoundError
from hydra_base import config
from hydra_base.util import dataset_util, get_json_as_string, get_json_as_dict
from hydra_base.lib import units
from hydra_base.util.permissions import required_perms

from hydra_base.lib.template.utils import check_dimension, get_attr_by_name_and_dimension

from hydra_base.lib.template.xml import (get_template_as_xml, import_template_xml,
    get_network_as_xml_template)

from hydra_base.lib.template.resource import (get_types_by_attr,
    apply_template_to_network,
    set_network_template,
    remove_template_from_network,
    get_matching_resource_types, assign_types_to_resources,
    check_type_compatibility,
    assign_type_to_resource,
    set_resource_type,
    get_network_template,
    remove_type_from_resource,
    validate_attr,
    validate_attrs,
    validate_scenario,
    validate_resource,
    validate_resourcescenario,
    validate_network)


from hydra_base.lib.cache import cache

log = logging.getLogger(__name__)

global CACHE_KEY
CACHE_KEY = 'template'

def _save_template_to_cache(template):
    cache.set(f"{CACHE_KEY}_{template.id}", template)

def _remove_template_from_cache(template_id):
    """
        If a template is in the cache, remove it.
    """
    cache.delete(f"{CACHE_KEY}_{template_id}")
    log.info("Template %s removed from cache.", template_id)

def parse_json_typeattr(type_i, typeattr_j, attribute_j, default_dataset_j, user_id=None):
    dimension_i = None
    if attribute_j.dimension_id is not None:
        # The dimension_id of the attribute is not None
        dimension_i = units.get_dimension(attribute_j.dimension_id)
    elif attribute_j.dimension is not None:
        # The dimension name of the attribute is not None
        dimension_name = attribute_j.dimension.strip()
        if dimension_name.lower() in ('dimensionless', ''):
            dimension_name = 'dimensionless'
        dimension_i = units.get_dimension_by_name(dimension_name.strip())
    elif attribute_j.unit_id is not None:
        # The unit_id of the attribute is not None
        dimension_i = units.get_dimension_by_unit_id(attribute_j.unit_id)
    elif attribute_j.unit not in ('', None):
        # The unit of the attribute is not None
        attribute_unit_id = units.get_unit_by_abbreviation(attribute_j.unit).id
        attribute_j.unit_id = attribute_unit_id
        dimension_i = units.get_dimension_by_unit_id(attribute_j.unit_id)

    attribute_name = attribute_j.name.strip()

    if dimension_i is None:
        # In this case we must get the attr with dimension id not set
        attr_i = get_attr_by_name_and_dimension(attribute_name, None)
    else:
        attr_i = get_attr_by_name_and_dimension(attribute_name, dimension_i.id)

    #Get an ID for the attribute
    db.DBSession.flush()

    for ta in type_i.typeattrs:
        if ta.attr_id == attr_i.id:
           typeattr_i = ta
           break
    else:
        typeattr_i = TypeAttr()
        log.debug("Creating type attr: type_id=%s, attr_id=%s", type_i.id, attr_i.id)
        typeattr_i.type_id = type_i.id
        typeattr_i.attr_id = attr_i.id
        typeattr_i.attr_is_var = typeattr_j.attr_is_var
        typeattr_i.attr = attr_i
        typeattr_i.status = 'A'
        type_i.typeattrs.append(typeattr_i)
        db.DBSession.add(typeattr_i)


    unit_id = None
    if attribute_j.unit_id is not None:
        typeattr_i.unit_id = typeattr_j.unit_id

    check_dimension(typeattr_i)

    if typeattr_j.description is not None:
        typeattr_i.description = typeattr_j.description

    if typeattr_j.properties is not None:
        if isinstance(typeattr_j.properties, dict):
            typeattr_i.properties = json.dumps(typeattr_j.properties)
        else:
            typeattr_i.properties = typeattr_j.properties

    if typeattr_j.is_var is not None:
        typeattr_i.attr_is_var = typeattr_j.is_var

    if typeattr_j.data_type is not None:
        typeattr_i.data_type = typeattr_j.data_type

    if default_dataset_j is not None:
        default = default_dataset_j

        unit = default.unit
        unit_id = None
        if unit not in (None, ''):
            unit_id = units.get_unit_by_abbreviation(unit).id

        if unit_id is None and typeattr_i.unit_id is not None:
            unit_id = typeattr_i.unit_id

        if unit_id is not None:
            check_dimension(typeattr_i, unit_id)

        if unit_id is not None and typeattr_i.unit_id is not None:
            if unit_id != typeattr_i.unit_id:
                raise HydraError("Default value has a unit of %s but the attribute"
                             " says the unit should be: %s"%(typeattr_i.unit_id, unit_id))

        val  = default.value

        data_type = default.type
        name = default.name if default.name not in (None, '') else "%s Default"%attr_i.name

        dataset_i = add_dataset(data_type,
                               val,
                               unit_id,
                               name= name,
                               user_id=user_id)
        typeattr_i.default_dataset_id = dataset_i.id


    if typeattr_j.restriction is not None or typeattr_j.data_restriction is not None:
        restriction = typeattr_j.restriction if typeattr_j.restriction is not None else typeattr_j.data_restriction
        if isinstance(restriction, dict):
            typeattr_i.data_restriction = json.dumps(restriction)
        else:
            typeattr_i.data_restriction = restriction
    else:
        typeattr_i.data_restriction = None

    return typeattr_i

@required_perms("get_template")
def get_template_as_json(template_id, **kwargs):
    """
        Get a template (including attribute and dataset definitions) as a JSON
        string. This is just a wrapper around the get_template_as_dict function.
    """
    user_id = kwargs['user_id']
    return json.dumps(get_template_as_dict(template_id, user_id=user_id))

@required_perms("get_template")
def get_template_as_dict(template_id, **kwargs):
    attr_dict = {}
    dataset_dict = {}

    template_i = db.DBSession.query(Template).filter(
                    Template.id==template_id).options(
                        joinedload(Template.templatetypes)\
                        .joinedload(TemplateType.typeattrs)\
                        .joinedload(TypeAttr.default_dataset)\
                        .joinedload(ModelDataset.metadata)
                    ).one()



    #Load all the attributes
    for type_i in template_i.templatetypes:
        for typeattr_i in type_i.typeattrs:
            typeattr_i.attr

    template_j = JSONObject(template_i)

    for tmpltype_j in template_j.templatetypes:
        ##Try to load the json into an object, as it will be re-encoded as json,
        ##and we don't want double encoding:
        if tmpltype_j.layout is not None:
            tmpltype_j.layout = get_json_as_dict(tmpltype_j.layout)

        for typeattr_j in tmpltype_j.typeattrs:
            typeattr_j.attr_id = str(typeattr_j.attr_id*-1)
            attr_dict[typeattr_j.attr_id] = JSONObject(
                                   {'name': typeattr_j.attr.name,
                                    'dimension_id':typeattr_j.attr.dimension_id
                                    })

            if typeattr_j.default_dataset_id is not None:
                typeattr_j.default_dataset_id = str(typeattr_j.default_dataset_id * -1)
                dataset_dict[typeattr_j.default_dataset_id] = JSONObject(
                            {
                                'name'     : typeattr_j.default_dataset.name,
                                'type'     : typeattr_j.default_dataset.type,
                                'unit_id'  : typeattr_j.default_dataset.unit_id,
                                'value'    : typeattr_j.default_dataset.value,
                                'metadata' : typeattr_j.default_dataset.metadata
                            })

            if hasattr(typeattr_j, 'default_dataset') and typeattr_j.default_dataset is not None:
                del(typeattr_j['default_dataset'])

            if hasattr(typeattr_j, 'attr') and typeattr_j.attr is not None:
                del(typeattr_j['attr'])

    output_data = {'attributes': attr_dict, 'datasets':dataset_dict, 'template': template_j}

    return output_data

@required_perms("add_template")
def import_template_json(template_json_string, allow_update=True, **kwargs):
    """
        Add the template, type and typeattrs described in a JSON file.

        Delete type, typeattr entries in the DB that are not in the JSON file.
        The assumption is that they have been deleted and are no longer required.

        The allow_update indicates whether an existing template of the same name should
        be updated, or whether it should throw an 'existing name' error.
    """

    user_id = kwargs.get('user_id')

    try:
        template_dict = json.loads(template_json_string)
    except:
        raise HydraError("Unable to parse JSON string. Plese ensure it is JSON compatible.")

    return import_template_dict(template_dict, allow_update=allow_update, user_id=user_id)

@required_perms("add_template")
def import_template_dict(template_dict, allow_update=True, **kwargs):

    user_id = kwargs.get('user_id')

    template_file_j = template_dict

    file_attributes = template_file_j.get('attributes')
    file_datasets   = template_file_j.get('datasets', {})
    template_j = JSONObject(template_file_j.get('template', {}))

    #default datasets are optional, so don't force them to exist in the structure
    default_datasets_j = {}
    for k, v in file_datasets.items():
        default_datasets_j[int(k)] = Dataset(v)

    if file_attributes is None or len(template_j) == 0:
        raise HydraError("Invalid template. The template must have the following structure: " + \
                            "{'attributes':\\{...\\}, 'datasets':\\{...\\}, 'template':\\{...\\}}")


    #Normalise attribute IDs so they're always ints (in case they're specified as strings)
    attributes_j = {}
    for k, v in file_attributes.items():
        attributes_j[int(k)] = JSONObject(v)

    template_name = template_j.name
    template_description = template_j.description

    template_layout = None
    if template_j.layout is not None:
        if isinstance(template_j.layout, dict):
            template_layout = json.dumps(template_j.layout)
        else:
            template_layout = template_j.layout

    try:
        template_i = db.DBSession.query(Template).filter(
            Template.name==template_name).options(
                joinedload(Template.templatetypes)
                .joinedload(TemplateType.typeattrs)
                .joinedload(TypeAttr.attr)).one()
        if allow_update == False:
            raise HydraError("Existing Template Found with name %s"%(template_name,))
        else:
            template_i.layout = template_layout
            template_i.description = template_description
    except NoResultFound:
        log.debug("Template not found. Creating new one. name=%s", template_name)
        template_i = Template(name=template_name, description=template_description, layout=template_layout)
        db.DBSession.add(template_i)

    types_j = template_j.templatetypes
    type_id_map = {r.id:r for r in template_i.templatetypes}
    #Delete any types which are in the DB but no longer in the JSON file
    type_name_map = {r.name:r.id for r in template_i.templatetypes}
    attr_name_map = {}
    for type_i in template_i.templatetypes:
        for typeattr in type_i.typeattrs:
            attr_name_map[typeattr.attr.name] = (typeattr.attr_id, typeattr.type_id)

    existing_types = set([r.name for r in template_i.templatetypes])
    log.debug(["%s : %s" %(tt.name, tt.id) for tt in template_i.templatetypes])
    log.debug("Existing types: %s", existing_types)

    new_types = set([t.name for t in types_j])
    log.debug("New Types: %s", new_types)

    types_to_delete = existing_types - new_types
    log.debug("Types to delete: %s", types_to_delete)
    log.debug(type_name_map)
    for type_to_delete in types_to_delete:
        type_id = type_name_map[type_to_delete]
        try:
            for i, tt in enumerate(template_i.templatetypes):
                if tt.id == type_id:
                    type_i = template_i.templatetypes[i]

                    #first remove all the type attributes associated to the type
                    for ta_i in type_i.typeattrs:
                        db.DBSession.delete(ta_i)

                    del(template_i.templatetypes[i])
                    log.debug("Deleting type %s (%s)", type_i.name, type_i.id)
                    del(type_name_map[type_to_delete])
                    db.DBSession.delete(type_i)
        except NoResultFound:
            pass


    #Add or update types.
    for type_j in types_j:
        type_name = type_j.name

        #check if the type is already in the DB. If not, create a new one.
        type_is_new = False
        if type_name in existing_types:
            type_id = type_name_map[type_name]
            type_i = type_id_map[type_id]
        else:
            log.debug("Type %s not found, creating new one.", type_name)
            type_i = TemplateType()
            type_i.name = type_name
            template_i.templatetypes.append(type_i)
            type_i.status = 'A' ## defaults to active
            type_is_new = True

        if type_j.description is not None:
            type_i.description = type_j.description

        if type_j.alias is not None:
            type_i.alias = type_j.alias

        #Allow 'type' or 'resource_type' to be accepted
        if type_j.type is not None:
            type_i.resource_type = type_j.type
        elif type_j.resource_type is not None:
            type_i.resource_type = type_j.resource_type

        if type_j.resource_type is None:
            raise HydraError("No resource type specified."
                             " 'NODE', 'LINK', 'GROUP' or 'NETWORK'")

        if type_j.layout is not None:
            if isinstance(type_j, dict):
                type_i.layout = json.dumps(type_j.layout)
            else:
                type_i.layout = type_j.layout

        #delete any TypeAttrs which are in the DB but not in the JSON file
        existing_attrs = []
        if not type_is_new:
            for r in template_i.templatetypes:
                if r.name == type_name:
                    for typeattr in r.typeattrs:
                        existing_attrs.append(typeattr.attr.name)

        existing_attrs = set(existing_attrs)

        type_attrs = []
        for typeattr_j in type_j.typeattrs:
            if typeattr_j.attr_id is not None:
                attr_j = attributes_j[typeattr_j.attr_id].name
            elif typeattr_j.attr is not None:
                attr_j = typeattr_j.attr.name
            type_attrs.append(attr_j)

        type_attrs = set(type_attrs)

        attrs_to_delete = existing_attrs - type_attrs
        for attr_to_delete in attrs_to_delete:
            attr_id, type_id = attr_name_map[attr_to_delete]
            try:
                attr_i = db.DBSession.query(TypeAttr).filter(TypeAttr.attr_id==attr_id, TypeAttr.type_id==type_id).options(joinedload(TypeAttr.attr)).one()
                db.DBSession.delete(attr_i)
                log.debug("Attr %s in type %s deleted",attr_i.attr.name, attr_i.templatetype.name)
            except NoResultFound:
                log.debug("Attr %s not found in type %s"%(attr_id, type_id))
                continue

        #Add or update type typeattrs
        #Support an external attribute dict or embedded attributes.
        for typeattr_j  in type_j.typeattrs:
            if typeattr_j.attr_id is not None:
                attr_j = attributes_j[typeattr_j.attr_id]
            elif typeattr_j.attr is not None:
                attr_j = typeattr_j.attr

            default_dataset_j = None
            if typeattr_j.default_dataset is not None:
                default_dataset_j = typeattr_j.default_dataset
            elif typeattr_j.default is not None: # for backward compatibility
                default_dataset_j = typeattr_j.default
            elif typeattr_j.default_dataset_id is not None:
                default_dataset_j = default_datasets_j[int(typeattr_j.default_dataset_id)]

            parse_json_typeattr(type_i, typeattr_j, attr_j, default_dataset_j, user_id=user_id)

    db.DBSession.flush()

    template_j = JSONObject(template_i)
    _save_template_to_cache(template_j)


    return template_i

def _parse_data_restriction(restriction_dict):
    if restriction_dict is None or len(restriction_dict) == 0:
        return None

    #replace soap text with an empty string
    #'{soap_server.hydra_complexmodels}' -> ''
    dict_str = re.sub('{[a-zA-Z._]*}', '', str(restriction_dict))

    if isinstance(restriction_dict, dict):
        new_dict = restriction_dict
    else:
        try:
            new_dict = json.loads(restriction_dict)
        except:
            raise HydraError(f"Unable to parse the JSON in the restriction data: {restriction_dict}")

    #Evaluate whether the dict actually contains anything.
    if not isinstance(new_dict, dict) or len(new_dict) == 0:
        log.critical('A restriction was specified, but it is null')
        return None

    ret_dict = {}
    for k, v in new_dict.items():
        if (isinstance(v, str) or isinstance(v, list)) and len(v) == 1:
            ret_dict[k] = v[0]
        else:
            ret_dict[k] = v

    return json.dumps(ret_dict)

@required_perms("add_template")
def add_template(template, **kwargs):
    """
        Add template and a type and typeattrs.
    """
    tmpl = Template()
    tmpl.name = template.name
    if template.parent_id:
        tmpl.parent_id = template.parent_id
    if template.description:
        tmpl.description = template.description
    if template.layout:
        tmpl.layout = get_json_as_string(template.layout)

    db.DBSession.add(tmpl)

    if template.templatetypes is not None:
        types = template.templatetypes
        for templatetype in types:
            ttype = _update_templatetype(templatetype, **kwargs)
            tmpl.templatetypes.append(ttype)

    db.DBSession.flush()

    log.info("[Added template]\n{}".format(template))

    template_j = JSONObject(tmpl)
    _save_template_to_cache(template_j)

    return tmpl


def _get_network_type(template_id):
    """find the network type in a template"""

    templatetypes_i = db.DBSession.query(TemplateType).filter(TemplateType.template_id == template_id).all()

    for templatetype in templatetypes_i:
        if templatetype.resource_type == 'NETWORK':
            return templatetype
    return None

def add_child_template(parent_id, name, description=None, **kwargs):
    """
        Add template and a type and typeattrs.
    """
    parent_template = db.DBSession.query(Template).filter(Template.id == parent_id).one()

    tmpl = Template()
    tmpl.name = name
    if description is not None:
        tmpl.description = description
    else:
        tmpl.description = parent_template.description
    tmpl.layout = parent_template.layout
    tmpl.parent_id = parent_id

    #now add a default network type, but only if the parent has one defined
    parent_type = _get_network_type(parent_id)

    if parent_type is not None:

        network_type = TemplateType()

        network_type.name = "{}-network".format(tmpl.name)
        network_type.resource_type = 'NETWORK'
        network_type.parent_id = parent_type.id
        network_type.status = 'A'

        tmpl.templatetypes.append(network_type)
    else:
        log.warn("Unable to set a network type on this template as its "
                 "parent %s does not contain one, so cannot create a "
                 "parent-child relationship.", parent_id)

    db.DBSession.add(tmpl)

    db.DBSession.flush()

    template_j = JSONObject(tmpl)

    #dont save this to the cache as it does not contain the full template information
    #wait for a get_template call to do this, to speed up user experience.

    return tmpl

def _set_template_status(template_id, status, **kwargs):
    """
        Set the status of a template to the specified status.
        These can be 'A' or 'X'
    """
    tmpl = db.DBSession.query(Template).filter(Template.id == template_id).one()

    tmpl.status = status

    _remove_template_from_cache(tmpl.id)

    db.DBSession.flush()

@required_perms("edit_template")
def activate_template(template_id, **kwargs):
    """
        Set the status of a template to active
    """

    _set_template_status(template_id, 'A')

@required_perms("edit_template")
def deactivate_template(template_id, **kwargs):
    """
        Set the status of a template to inactive
    """

    _set_template_status(template_id, 'X')

@required_perms("edit_template")
def update_template(template, auto_delete=False, **kwargs):
    """
        Update template and a type and typeattrs.
        args:
            template (JSONObject): The template to update
            auto_delete (bool): A flag to indicated whether missing types from `template.templatetypes`
                                should be deleted automatically. This flag is also
                                used when updating the typeattrs of type. Defaults to False.
    """
    tmpl = db.DBSession.query(Template).filter(Template.id == template.id).one()
    tmpl.name = template.name

    if template.status is not None:
        tmpl.status = template.status

    if template.description:
        tmpl.description = template.description

    template_types = tmpl.get_types()

    if template.layout:
        tmpl.layout = get_json_as_string(template.layout)

    type_dict = dict([(t.id, t) for t in template_types])

    #a list of all the templatetypes in the incoming template object
    req_templatetype_ids = []

    if template.types is not None or template.templatetypes is not None:
        types = template.types if template.types is not None else template.templatetypes
        for templatetype in types:

            if templatetype.id is not None and templatetype.template_id != tmpl.id:
                log.debug("Type %s is a part of a parent template. Ignoring.", templatetype.id)
                req_templatetype_ids.append(type_i.id)
                continue

            if templatetype.id is not None:
                type_i = type_dict[templatetype.id]
                _update_templatetype(templatetype, auto_delete=auto_delete, **kwargs)
                req_templatetype_ids.append(type_i.id)
            else:
                #Give it a template ID if it doesn't have one
                templatetype.template_id = template.id
                new_templatetype_i = _update_templatetype(templatetype, auto_delete=auto_delete, **kwargs)
                req_templatetype_ids.append(new_templatetype_i.id)

    if auto_delete is True:
        for ttype in template_types:
            if ttype.id not in req_templatetype_ids:
                delete_templatetype(ttype.id, **kwargs)

    db.DBSession.flush()

    updated_templatetypes = tmpl.get_types()

    tmpl_j = JSONObject(tmpl)

    tmpl_j.templatetypes = updated_templatetypes

    _save_template_to_cache(tmpl_j)

    return tmpl_j

@required_perms("delete_template")
def delete_template(template_id, delete_resourcetypes=False, **kwargs):
    """
        Delete a template and its type and typeattrs.
        The 'delete_resourcetypes' flag forces the template to remove any resource types
        associated to the types in the network. Use with caution!
    """
    try:
        tmpl = db.DBSession.query(Template).filter(Template.id == template_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Template %s not found"%(template_id,))

    for templatetype in tmpl.templatetypes:
        delete_templatetype(templatetype.id, flush=False, delete_resourcetypes=delete_resourcetypes, user_id=kwargs.get('user_id'))

    db.DBSession.delete(tmpl)
    db.DBSession.flush()

    _remove_template_from_cache(tmpl.id)

    return 'OK'

@required_perms("get_template")
def get_templates(load_all=True, include_inactive=False, **kwargs):
    """
        Get all templates.
        Args:
            load_all Boolean: Returns just the template entry or the full
            template structure (template types and type attrs)
            include_inactive Boolean: If true, returns all templates. If false, returns
            only templates with a status of 'A'
        Returns:
            List of Template objects
    """
    templates_i = db.DBSession.query(Template).options(joinedload(Template.templatetypes)).all()
    if load_all is True:
        full_templates = []
        for template_i in templates_i:
            full_template = get_template(template_i.id, **kwargs)
            full_templates.append(full_template)
    else:
        full_templates = [JSONObject(template_i) for template_i in templates_i]


    #Filter out all the inactive templates
    if include_inactive is False:
        full_templates = list(filter(lambda x:x.status == 'A', full_templates))

    return full_templates

@required_perms("edit_template")
def remove_attr_from_type(type_id, attr_id, **kwargs):
    """
        Remove an attribute from a type
    """
    typeattr_i = db.DBSession.query(TypeAttr).filter(TypeAttr.type_id == type_id,
                                                     TypeAttr.attr_id == attr_id).one()

    _remove_template_from_cache(typeattr_i.templatetype.template_id)

    db.DBSession.delete(typeattr_i)

@required_perms("get_template")
def get_template(template_id, **kwargs):
    """
        Get a specific resource template, by ID.
    """
    log.info("Getting template %s", template_id)
    tmpl_j = cache.get(f"{CACHE_KEY}_{template_id}")

    if tmpl_j is not None:
        log.info("Returning cached template")
        return tmpl_j

    try:
        log.info("Building template")

        tmpl_i = db.DBSession.query(Template).filter(
            Template.id == template_id).one()


        tmpl_j = JSONObject(tmpl_i)

        tmpl_j.templatetypes = tmpl_i.get_types()

        #ignore the messing around we've been doing to the ORM objects
        #db.DBSession.expunge(tmpl_i)

        _save_template_to_cache(tmpl_j)


        return tmpl_j
    except NoResultFound:
        raise HydraError("Template %s not found"%template_id)

@required_perms("get_template")
def get_template_by_name(name, **kwargs):
    """
        Get a specific resource template, by name.
    """
    try:
        tmpl_i = db.DBSession.query(Template).filter(
            Template.name == name).one()

        tmpl_j = get_template(tmpl_i.id, **kwargs)

        return tmpl_j

    except NoResultFound:
        log.info("%s is not a valid identifier for a template", name)
        raise HydraError('Template "%s" not found'%name)

@required_perms("edit_template")
def add_templatetype(templatetype, **kwargs):
    """
        Add a template type with typeattrs.
    """

    type_i = _update_templatetype(templatetype, **kwargs)

    _remove_template_from_cache(type_i.template_id)

    #remove any child templates from the cache too
    child_types = db.DBSession.query(TemplateType).filter(TemplateType.parent_id == type_i.id).all()
    for child_type in child_types:
        _remove_template_from_cache(child_type.template_id)

    db.DBSession.flush()

    return type_i

def add_child_templatetype(parent_id, child_template_id, **kwargs):
    """
        Add a child templatetype
    """
    #check if the type is already there:
    #this means we can only add one child type per template
    existing_child = db.DBSession.query(TemplateType).filter(
        TemplateType.parent_id == parent_id,
        TemplateType.template_id == child_template_id).first()

    if existing_child is not None:
        return existing_child

    #Now check that we're not adding a child template type to a template type
    #which is already a child i.e. the template ID of the proposed parent is
    #the same as the one we want to put the child into
    parent_type = db.DBSession.query(TemplateType).filter(
        TemplateType.id == parent_id).one()

    if parent_type.template_id == child_template_id:
        return parent_type

    #The child doesn't exist already, so create it.
    child_type_i = TemplateType()
    child_type_i.template_id = child_template_id
    child_type_i.parent_id = parent_id

    db.DBSession.add(child_type_i)

    db.DBSession.flush()

    _remove_template_from_cache(child_template_id)

    return child_type_i

def add_child_typeattr(parent_id, child_template_id, **kwargs):
    """
        Add template and a type and typeattrs.
    """

    #does this child already exist in this template?
    existing_child_typeattr = db.DBSession.query(TypeAttr).join(TemplateType).filter(
        TypeAttr.parent_id == parent_id).filter(
            TemplateType.template_id == child_template_id).first()

    if existing_child_typeattr is not None:
        return existing_child_typeattr

    parent_typeattr = db.DBSession.query(TypeAttr)\
            .filter(TypeAttr.id == parent_id).one()

    child_type = add_child_templatetype(parent_typeattr.type_id, child_template_id)

    child_typeattr_i = TypeAttr()
    child_typeattr_i.attr_id = parent_typeattr.attr_id
    child_typeattr_i.type_id = child_type.id
    child_typeattr_i.parent_id = parent_id

    db.DBSession.add(child_typeattr_i)

    db.DBSession.flush()

    _remove_template_from_cache(child_template_id)

    return child_typeattr_i



@required_perms("edit_template")
def update_templatetype(templatetype, auto_delete=False, **kwargs):
    """
        Update a resource type and its typeattrs.
        New typeattrs will be added. typeattrs not sent will be ignored.
        To delete typeattrs, call delete_typeattr

        args:
            templatetype: A template type JSON object
            auto_delete (bool): Flag to indicate whether non-presence of
                                typeattrs in incoming object should delete them.
                                Default to False
    """

    tmpltype_i = db.DBSession.query(TemplateType).filter(TemplateType.id == templatetype.id).one()

    _update_templatetype(templatetype, tmpltype_i, auto_delete=auto_delete, **kwargs)

    db.DBSession.flush()

    updated_type = tmpltype_i.template.get_type(tmpltype_i.id)

#    db.DBSession.expunge(updated_type)

    _remove_template_from_cache(tmpltype_i.template_id)

    return updated_type

def _set_typeattr(typeattr, existing_ta=None):
    """
        Add or update a type attribute.
        If an existing type attribute is provided, then update.

        Checks are performed to ensure that the dimension provided on the
        type attr (not updateable) is the same as that on the referring attribute.
        The unit provided (stored on tattr) must conform to the dimension stored
        on the referring attribute (stored on tattr).

        This is done so that multiple templates can all use the same attribute,
        but specify different units.

        If no attr_id is provided, but an attr_name and dimension are provided,
        then a new attribute can be created (or retrieved) and used. I.e., no
        attribute ID must be specified if attr_name and dimension are specified.

        ***WARNING***
        Setting ID to null means a new type attribute (and even a new attr)
        may be added, None are removed or replaced. To remove other type attrs, do it
        manually using delete_typeattr
    """
    if existing_ta is None:

        #check for an existing TA
        check_existing_ta = db.DBSession.query(TypeAttr)\
            .filter(TypeAttr.attr_id == typeattr.attr_id, TypeAttr.type_id == typeattr.type_id).first()

        #There's already a TA with this attr_id in this type
        if check_existing_ta is not None:
            ta = check_existing_ta
        else:
            ta = TypeAttr(attr_id=typeattr.attr_id)
            ## default new type attrs to 'active'.
            ##This has replaced the database default because for child typeattrs,
            ##we need the status to be NULL so it can inherit from its parent
            ta.status = 'A'
    else:
        if typeattr.id is not None:
            ta = db.DBSession.query(TypeAttr).filter(TypeAttr.id == typeattr.id).one()
        else:
            ta = existing_ta

    ta.attr_id = typeattr.attr_id
    ta.unit_id = typeattr.unit_id
    ta.type_id = typeattr.type_id
    ta.data_type = typeattr.data_type
    ta.status = typeattr.status if typeattr.status is not None else 'A'

    if hasattr(typeattr, 'default_dataset_id') and typeattr.default_dataset_id is not None:
        ta.default_dataset_id = typeattr.default_dataset_id

    ta.description = typeattr.description

    ta.properties = typeattr.get_properties()

    #support legacy use of 'is_var' instead of 'attr_is_var'
    if hasattr(typeattr, 'is_var') and typeattr.is_var is not None:
        typeattr.attr_is_var = typeattr.is_var

    ta.attr_is_var = typeattr.attr_is_var if typeattr.attr_is_var is not None else 'N'

    ta.data_restriction = _parse_data_restriction(typeattr.data_restriction)

    if typeattr.unit_id is None or typeattr.unit_id == '':
        # All right. Check passed
        ta.unit_id = None
        pass
    else:
        unit = units.get_unit(typeattr.unit_id)
        dimension = units.get_dimension(unit.dimension_id)
        if typeattr.attr_id is not None and typeattr.attr_id > 0:
            # Getting the passed attribute, so we need to check consistency
            # between attr dimension id and typeattr dimension id
            attr = db.DBSession.query(Attr).filter(Attr.id==ta.attr_id).first()
            if attr is not None and attr.dimension_id is not None and\
               attr.dimension_id != dimension.id or \
               attr is not None and attr.dimension_id is None:

                attr_dimension = units.get_dimension(attr.dimension_id)
                # In this case there is an inconsistency between
                # attr.dimension_id and typeattr.unit_id
                raise HydraError("Unit mismatch between type and attirbute."+
                                 f"Type attribute for {attr.name} specifies "+
                                 f"unit {unit.name}, dimension {dimension.name}."+
                                 f"The attribute specifies a dimension of {attr_dimension.name}"+
                                 "Cannot set a unit on a type attribute which "+
                                 "does not match its attribute.")
        elif typeattr.attr_id is None and typeattr.name is not None:
            # Getting/creating the attribute by typeattr dimension id and typeattr name
            # In this case the dimension_id "null"/"not null" status is ininfluent
            attr = get_attr_by_name_and_dimension(typeattr.name, dimension.id)

            ta.attr_id = attr.id
            ta.attr = attr

    check_dimension(ta)

    if existing_ta is None:
        log.debug("Adding ta to DB")
        db.DBSession.add(ta)

    attr = db.DBSession.query(Attr).filter(Attr.id == ta.attr_id).one()
    ta.attr = attr

    return ta

def _set_cols(source, target, reference=None):
    """
        Set the value on the column of a target row.
        This checks if the value is the same as that of another reference row,
        and if it is the same, it sets that column to None.
        Ex:
            reference is: {'status': 'x'} and value is 'x'}, then
            target will be set to {'status': None}. If status is 'y',
            then target will be {'status', 'y'}
        Args:
            target: DB row to write the column to
            source: The incoming object (a JSONObject) containing the request data
            reference: DB row used for comparison
            colnames: The column names to set
            value: The value to set.
    """
    for colname in source:

        if colname not in target.__table__.columns:
            continue

        if colname in ['cr_date', 'updated_at']:
            continue

        if hasattr(reference, '_protected_columns')\
           and colname in reference._protected_columns:
            #as a child, you can't change stuff like IDS, cr dates etc.
            continue

        if target.parent_id is not None\
           and hasattr(reference, '_hierarchy_columns')\
           and colname in reference._hierarchy_columns:
            #as a child, you can't change stuff like IDS, cr dates etc.
            continue

        newval = getattr(source, colname)

        if colname == 'layout':
            newval = get_json_as_string(newval)

        if reference is None:
            setattr(target, colname, newval)
            continue

        refval = getattr(reference, colname)

        if newval != refval:
            setattr(target, colname, newval)


def _update_templatetype(templatetype, existing_tt=None, auto_delete=False, **kwargs):
    """
        Add or update a templatetype. If an existing template type is passed in,
        update that one. Otherwise search for an existing one. If not found, add.
    """
    #flag to indicate if this update results in an insert
    is_new = False

    if existing_tt is None:
        if "id" in templatetype and templatetype.id is not None:
            tmpltype_i = db.DBSession.query(TemplateType).filter(
                TemplateType.id == templatetype.id).one()
        else:
            is_new = True
            tmpltype_i = TemplateType()
            tmpltype_i.template_id = templatetype.template_id

            ## default new template types to active
            ## This has replaced the database default because for child typeattrs,
            ## we need the status to be NULL so it can inherit from its parent
            tmpltype_i.status = 'A'
    else:
        tmpltype_i = existing_tt

    _set_cols(templatetype, tmpltype_i, existing_tt)

    ta_dict = {}
    for t in tmpltype_i.get_typeattrs():
        ta_dict[t.attr_id] = t

    existing_attrs = []

    if templatetype.typeattrs is not None:
        for typeattr in templatetype.typeattrs:
            if typeattr.attr_id in ta_dict:
                #this belongs to a parent. Ignore.
                if typeattr.type_id is not None and typeattr.type_id != tmpltype_i.id:
                    continue

                ta = _set_typeattr(typeattr, ta_dict[typeattr.attr_id])
                existing_attrs.append(ta.attr_id)
            else:
                ta = _set_typeattr(typeattr)
                tmpltype_i.typeattrs.append(ta)
                existing_attrs.append(ta.attr_id)

    log.debug("Deleting any type attrs not sent")
    if auto_delete is True:
        for ta in ta_dict.values():
            if ta.attr_id not in existing_attrs:
                delete_typeattr(ta, **kwargs)

    if is_new is True:
        db.DBSession.add(tmpltype_i)

    return tmpltype_i

@required_perms("edit_template")
def delete_templatetype(type_id, template_i=None, delete_resourcetypes=False, flush=True, delete_children=False, **kwargs):
    """
        Delete a template type and its typeattrs.
    """

    try:
        tmpltype_i = db.DBSession.query(TemplateType)\
                .filter(TemplateType.id == type_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Template Type %s not found"%(type_id,))

    if template_i is None:
        template_i = db.DBSession.query(Template).filter(
            Template.id == tmpltype_i.template_id).one()

    if len(tmpltype_i.get_children()) > 0 and delete_children is False:
        raise HydraError(f"Unable to delete type. Template type {tmpltype_i.name} (ID: {type_id}) has"
                         f"children. If you want to delete this, use the 'delete_children' flag.")

    if delete_children is True:
        tmpltype_i.delete_children(delete_resourcetypes=delete_resourcetypes)

    tmpltype_i.check_can_delete_resourcetypes(delete_resourcetypes=delete_resourcetypes)

    if delete_resourcetypes is True:
        tmpltype_i.delete_resourcetypes()

    #first remove the templatetypes
    for ta_i in tmpltype_i.typeattrs:
        db.DBSession.delete(ta_i)

    db.DBSession.delete(tmpltype_i)

    _remove_template_from_cache(tmpltype_i.template_id)

    if flush:
        db.DBSession.flush()

@required_perms("get_template")
def get_templatetype(type_id, include_parent_data=True, **kwargs):
    """
        Get a specific template type by ID. As types can be inherited, this
        type may contain data from its parent. If the 'include_parent_type' is false,
        then just the data for this templatetype is returned.
    """

    #First get the DB entry
    templatetype = db.DBSession.query(TemplateType).filter(
        TemplateType.id == type_id).options(noload(TemplateType.typeattrs)).one()

    if include_parent_data is False:
        return templatetype

    #then get the template
    template = db.DBSession.query(Template).filter(
        Template.id == templatetype.template_id).one()

    #then get the type, but this time with inherited data.
    inherited_templatetype = template.get_type(type_id)

    #ignore the messing around we've been doing to the ORM objects
#    db.DBSession.expunge(inherited_templatetype)

    return inherited_templatetype

@required_perms("edit_template")
def clone_templatetype(type_id, name=None, **kwargs):
    """
        Clone a specific template type by ID. Creates a copy of the template type
        with all its typeattrs. The clone will have new IDs and timestamps but
        identical content otherwise, except the name which will be modified to be unique.
    """

    #First get the DB entry
    parent_templatetype = db.DBSession.query(TemplateType).filter(
        TemplateType.id == type_id).options(noload(TemplateType.typeattrs)).one()

    cloned_templatetype = TemplateType()
    for col in parent_templatetype.__table__.columns:
        if col.name in ['id', 'cr_date', 'updated_at']:
            continue
        setattr(cloned_templatetype, col.name, getattr(parent_templatetype, col.name))

    if name is not None:
        base_name = name
    else:
        # Make the name unique by appending " (Clone)" and potentially a number
        base_name = parent_templatetype.name + " (Clone)"
    clone_name = base_name
    counter = 1
    # Check if a template type with this name already exists in the same template
    while db.DBSession.query(TemplateType).filter(
        TemplateType.template_id == parent_templatetype.template_id,
        TemplateType.name == clone_name,
        TemplateType.resource_type == parent_templatetype.resource_type
    ).first() is not None:
        clone_name = f"{base_name} {counter}"
        counter += 1

    cloned_templatetype.name = clone_name

    typeattrs = db.DBSession.query(TypeAttr).filter(
        TypeAttr.type_id == parent_templatetype.id).all()
    for typeattr in typeattrs:
        cloned_typeattr = TypeAttr()
        for col in typeattr.__table__.columns:
            if col.name in ['id', 'cr_date', 'updated_at']:
                continue
            setattr(cloned_typeattr, col.name, getattr(typeattr, col.name))
        cloned_templatetype.typeattrs.append(cloned_typeattr)

    db.DBSession.add(cloned_templatetype)
    db.DBSession.flush()  # Flush to get the ID

    return cloned_templatetype

@required_perms("get_template")
def get_typeattr(typeattr_id, include_parent_data=True, **kwargs):
    """
        Get a specific resource type by ID.
    """

    typeattr = db.DBSession.query(TypeAttr)\
            .filter(TypeAttr.id == typeattr_id)\
            .options(joinedload(TypeAttr.default_dataset)).one()

    if include_parent_data is False:
        return typeattr

    template = typeattr.templatetype.template

    #then get the type, but this time with inherited data.
    inherited_typeattr = template.get_typeattr(typeattr_id)

#    db.DBSession.expunge(inherited_typeattr)

    return inherited_typeattr



@required_perms("get_template")
def get_templatetype_by_name(template_id, type_name, **kwargs):
    """
        Get a specific resource type by name.
    """

    try:
        templatetype = db.DBSession.query(TemplateType).filter(
            TemplateType.template_id == template_id,
            TemplateType.name == type_name).one()
    except NoResultFound:
        raise HydraError("%s is not a valid identifier for a type"%(type_name))

    inherited_templatetype = get_templatetype(templatetype.id, **kwargs)

    return inherited_templatetype

@required_perms("edit_template")
def add_typeattr(typeattr, **kwargs):
    """
        Add an typeattr to an existing type.
    """

    tmpltype = get_templatetype(typeattr.type_id, user_id=kwargs.get('user_id'))

    ta = _set_typeattr(typeattr)

    _remove_template_from_cache(tmpltype.template_id)

    db.DBSession.flush()

    return ta

@required_perms("edit_template")
def update_typeattr(typeattr, **kwargs):
    """
        Update an existing an typeattr with updated values.
    """
    #First check if an existing one already exists.
    existing_ta = None
    if typeattr.id is not None:
        existing_ta = db.DBSession.query(TypeAttr)\
                .filter(TypeAttr.id == typeattr.id).first()
    else:
        existing_ta = db.DBSession.query(TypeAttr)\
            .filter(TypeAttr.type_id == typeattr.type_id,
                    TypeAttr.attr_id == typeattr.attr_id).first()

    typeattr_updated = _set_typeattr(typeattr, existing_ta)

    db.DBSession.flush()

    _remove_template_from_cache(existing_ta.templatetype.template_id)

    return typeattr_updated

@required_perms("edit_template")
def delete_typeattr(typeattr_id, **kwargs):
    """
        Remove an typeattr from an existing type
    """

    ta = db.DBSession.query(TypeAttr).filter(TypeAttr.id == typeattr_id).one()

    db.DBSession.delete(ta)

    db.DBSession.flush()

    _remove_template_from_cache(ta.templatetype.template_id)

    return 'OK'

@required_perms("get_template")
def get_all_parent_types(ttype_id, **kwargs):
    """
        Returns all TemplateTypes which are parents of
        the TemplateType with <ttype_id>
    """
    parent_ids = Template.get_type_parent_ids(ttype_id)

    tt_qry = db.DBSession.query(TemplateType)
    tt_qry = tt_qry.filter(TemplateType.id.in_(parent_ids))
    tt_qry = tt_qry.options(noload(TemplateType.typeattrs))
    return tt_qry.all()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from sqlalchemy import create_engine,\
        MetaData,\
        Table
from sqlalchemy.engine import reflection
from sqlalchemy.orm import sessionmaker
import logging
from mysql.connector.connection import MySQLConnection
from .. import config
import os
import datetime
from multiprocessing import Process
from sqlalchemy.ext.declarative import declarative_base
from decimal import Decimal
from sqlalchemy import and_

Base = declarative_base()


engine_name = config.get('mysqld', 'url')

def connect_mysql():
    """
        return an inspector object
    """
    MySQLConnection.get_characterset_info = MySQLConnection.get_charset

    db = create_engine(engine_name)
    db.echo = True
    db.connect()
    
    return db

sqlite_engine = "sqlite:///%s"%(config.get('sqlite', 'backup_url'))

def connect_sqlite():

    db = create_engine(sqlite_engine, echo=True)
    db.connect()
    return db

def truncate_all_audit_tables():
    logging.info("Starting truncation")

    mysql_db = connect_mysql()
    sqlite_db = connect_sqlite()

    # create a configured "Session" class
    session1 = sessionmaker(bind=mysql_db)
    session2 = sessionmaker(bind=sqlite_db)

    # create a Session
    mysql_session = session1()
    sqlite_session = session2() 

    mysql_metadata = MetaData(mysql_db)
    sqlite_metadata = MetaData(sqlite_db)

    insp = reflection.Inspector.from_engine(mysql_db)

    tables = {}
    audit_tables = {}
    for table_name in insp.get_table_names():
        table = Table(table_name, mysql_metadata, autoload=True)
        if table_name.endswith('_aud'):
            audit_tables[table_name] = table
        else:
            tables[table_name] = table

    for t in tables:
       table       = tables[t]
       audit_table = audit_tables.get("%s_aud"%(t))
       if audit_table is None:
           logging.warning("Audit table for %s does not exist"%(table.name))
           continue
       export_table_to_sqlite(mysql_session, sqlite_session, sqlite_metadata, audit_table)
       truncate_table(mysql_session, table, audit_table)
    
    mysql_session.commit()
    sqlite_session.commit()
    logging.info("Truncation Complete")

def truncate_table(session, table, audit_table):
    """
        @args: session so queries can be made, table so primary key columns
        can be determined and the audit_table to be truncated.
    """
    pk_col_names = [x.name for x in table.primary_key.columns]

    truncated_cols = []
    for audit_col in audit_table.columns:
        if audit_col.name in pk_col_names:
            truncated_cols.append(audit_col)

    rs = session.query(*truncated_cols).group_by(*truncated_cols).all()

    for r in rs:
        args = []
        for arg in truncated_cols:
            args.append(arg == getattr(r, arg.name))
        
        aud_id_col = audit_table.c['aud_id']
        rs = session.query(aud_id_col).filter(and_(*args)).order_by(aud_id_col.desc())[3:]
        aud_ids = [str(r.aud_id) for r in rs]
        if len(aud_ids) > 0:
            delete_sql = "delete from %(table_name)s where aud_id in %(aud_ids)s;"\
                    %{'table_name':audit_table.name,
                      'aud_ids'   : "(%s)"%(','.join(aud_ids),)}
            session.execute(delete_sql)

def export_table_to_csv(session, table, target=None):
    """
        @args: session so queries can be made, table so primary key columns
        can be determined and the audit_table to be truncated.
    """

    if target is None:
        target_dir = os.path.join(config.get('db', 'export_target'))
        target = os.path.join(config.get('db', 'export_target'), table.name)

    if not os.path.exists(target_dir):
        os.mkdir(target_dir)

    if os.path.exists(target):
        target_file      = open(target, 'r+')
    
        rs = session.query(table).all()
        
        entries_in_db = set()
        for r in rs:
            entries_in_db.add("%s"%(r.__repr__()))

        contents = set(target_file.read().split('\n'))
        
        new_data = entries_in_db.difference(contents)

        if len(new_data) > 0:
            target_file.write("%s\n"%(datetime.datetime.now()))
            target_file.write("%s\n"%('.'.join([c.name for c in table.columns])))
            for d in new_data:
                target_file.write("%s\n"%(d))
    else:
        target_file      = open(target, 'w')
        rs = session.query(table).all()
        target_file.write("%s\n"%(datetime.datetime.now()))
        target_file.write("%s\n"%('.'.join([c.name for c in table.columns])))
        for r in rs:
            target_file.write("%s\n"%(r.__repr__()))

def export_table_to_sqlite(mysql_session, sqlite_session, sqlite_metadata, audit_table):
    """
        @args: session so queries can be made, table so primary key columns
        can be determined and the audit_table to be truncated.
    """


    current_data = mysql_session.query(audit_table).all()

    sqlite_table = Table(audit_table.name, sqlite_metadata, autoload=True)

    sqlite_data  =  sqlite_session.query(sqlite_table).all()

    entries_in_mysql_db = set(current_data)
    entries_in_sqlite_db = set(sqlite_data)
        
    new_data = list(entries_in_mysql_db.difference(entries_in_sqlite_db))
    if len(new_data) > 0:
        values = [] 
        for val in new_data:
            row = []
            for i, v in enumerate(val):
                if type(v) == Decimal:
                    row.append(float(v))
                else:
                    row.append(v)
            values.append(tuple(row))

        markers = ','.join('?' * len(new_data[0]))
        colnames = '(%s)'%(','.join(new_data[0].keys()))
        ins = 'INSERT INTO {tablename} {colnames} VALUES ({markers})'
        ins = ins.format(tablename=sqlite_table.name, colnames=colnames, markers=markers)
        res = sqlite_session.connection().execute(ins, values)

        logging.info(res)

def run():
    p = Process(target=truncate_all_audit_tables)
    p.start()
    #p.join()

if __name__ == '__main__':
    logging.basicConfig(level='INFO')
    truncate_all_audit_tables()




"""
Package for DatasetManagers and StorageAdapters.

"""

""" Dataset managers """
from .mongodatasetmanager import MongoDatasetManager

""" Storage adapters """
from .mongostorageadapter import MongoStorageAdapter
from .hdfstorageadapter import HdfStorageAdapter



"""dataset_collection

Revision ID: a48ad41579de
Revises: 08f0ebe40290
Create Date: 2018-04-17 14:43:11.872611

"""
from alembic import op
import sqlalchemy as sa

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'a48ad41579de'
down_revision = '08f0ebe40290'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tDatasetCollection

        try:
            op.alter_column('tDatasetCollection', 'collection_id', new_column_name='id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tDatasetCollection', 'collection_name', new_column_name='name', existing_type=sa.String(60), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tDatasetCollection
        try:
            op.drop_table('tDatasetCollection_new')
        except:
            log.info("tDatasetCollection_new isn't there")

        try:
            # ## tDatasetCollection
            op.create_table(
                'tDatasetCollection_new',
                sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('name', sa.Text(60), nullable=False),
                sa.Column('cr_date', sa.TIMESTAMP(), nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP'))
            )

            op.execute("insert into tDatasetCollection_new (id, name, cr_date) select collection_id, collection_name, cr_date from tDatasetCollection")

            op.rename_table('tDatasetCollection','tDatasetCollection_old')
            op.rename_table('tDatasetCollection_new', 'tDatasetCollection')
            op.drop_table('tDatasetCollection_old')

        except Exception as e:
            log.exception(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tDatasetCollection', 'id', new_column_name='collection_id', existing_type=sa.Integer(), primary_key=True, autoincrement=True, nullable=False)
            op.alter_column('tDatasetCollection', 'name', new_column_name='collection_name', existing_type=sa.String(60), nullable=False)
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tDatasetCollection
        try:
            op.drop_table('tDatasetCollection_new')
        except:
            log.info("tDatasetCollection_new isn't there")

        try:
            # ## tDatasetCollection
            op.create_table(
                'tDatasetCollection_new',
                sa.Column('collection_id', sa.Integer(), primary_key=True, nullable=False),
                sa.Column('collection_name', sa.Text(60), nullable=False),
                sa.Column('cr_date', sa.TIMESTAMP(), nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP'))
            )

            op.execute("insert into tDatasetCollection_new (collection_id, collection_name, cr_date) select id, name, cr_date from tDatasetCollection")

            op.rename_table('tDatasetCollection','tDatasetCollection_old')
            op.rename_table('tDatasetCollection_new', 'tDatasetCollection')
            op.drop_table('tDatasetCollection_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
"""
    NOTE:
    All the methods that has SKELETON key in the doc, are not yet implemented partially/fully

"""
# Example of a SKELETON METHOD
# def test_(self, client):
#     """
#         SKELETON
#     """
#     pass


import logging
import copy
import json
import hydra_base as hb
from hydra_base.exceptions import HydraError
import datetime
import pytest
from hydra_base.lib.objects import JSONObject
log = logging.getLogger(__name__)

class TestAttribute:
    """
        Test for attribute-based functionality
    """

    """
        TESTED
    """

    def test_add_attribute(self, client):
        test_attr = JSONObject({
            "name": f'Test Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        previous_all_attributes = client.get_attributes()

        new_attr = client.add_attribute(test_attr)

        assert new_attr.name == test_attr.name, \
            "add_attribute didn't work"

        all_attributes = client.get_attributes()

        assert len(all_attributes) == len(previous_all_attributes) + 1

        #try add it again. SHould have no effect
        new_attr = client.add_attribute(test_attr)

        all_attributes_insert_2 = client.get_attributes()

        assert len(all_attributes) == len(all_attributes_insert_2)

        upper_case_attr = test_attr
        upper_case_attr['name'] = upper_case_attr['name'].upper()

        new_upper_attr = client.add_attribute(upper_case_attr)

        assert new_upper_attr['id'] == new_attr['id']


    def test_update_attribute(self, client):
        test_attr = JSONObject({
            "name": f'Test Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        new_attr = client.add_attribute(test_attr)
        new_attr.name = f"Test attr updated {datetime.datetime.now()}"
        updated_attr = client.update_attribute(new_attr)
        assert new_attr.id == updated_attr.id and \
                updated_attr.name == new_attr.name, \
                "update_attribute didn't work"
        #Try update this again (should have no effect)
        updated_attr.description = "Updated description"
        updated_attr_1 = client.update_attribute(updated_attr)
        assert updated_attr_1.description == "Updated description"

        #Now add a new attribute which should fail when updated
        test_attr_fail = JSONObject({
            "name": f'Test Attribute {datetime.datetime.now()}',
            "dimension_id": None
        })

        new_attr_fail = client.add_attribute(test_attr_fail)
        #set the name to be the same as the other attribute
        new_attr_fail.name = new_attr.name

        #this should fail because there's already an attribute with this naem
        #and dimension (since we've just set it.)
        with pytest.raises(HydraError):
            client.update_attribute(new_attr_fail)



    def test_delete_attribute(self, client):
        test_attr = JSONObject({
            "name": 'Test Attribute 1',
            "dimension_id": None
        })
        new_attr = client.add_attribute(test_attr)

        attr = client.get_attribute_by_id(new_attr.id)

        result = client.delete_attribute(new_attr.id)

        with pytest.raises(HydraError):
            client.get_attribute_by_id(new_attr.id)

        with pytest.raises(HydraError):
            result = client.delete_attribute(new_attr.id)

    def test_add_attributes(self, client):
        test_attrs = [
            JSONObject({
                "name": 'Test Attribute 1',
                "dimension_id": None
            }),
            JSONObject({
                "name": 'Test Attribute 2',
                "dimension_id": 1
            })
        ]
        new_attrs_list_1 = client.add_attributes(test_attrs)

        all_attributes_after_add_1 = client.get_attributes()

        #test to make sure that the insert is not case-sensitive.
        upper_test_attrs = test_attrs
        for a in upper_test_attrs:
            a['name'] = a['name'].upper()

        #Try adding the attributes again. It should ignore them as theyr'e already there.
        new_attrs_list_2 = client.add_attributes(upper_test_attrs)

        all_attributes_after_add_2 = client.get_attributes()

        #This should have returned the attributes with the IDS from the first insert
        assert sorted([a.id for a in new_attrs_list_1]) == sorted([a.id for a in new_attrs_list_2])

        assert len(all_attributes_after_add_1) == len(all_attributes_after_add_2)

        attributeset = set([(a.name, a.dimension) for a in all_attributes_after_add_1])

        #Ensure that there are no duplicates by checking that the length of the set
        #of name/dimension pairs is the same as the length of all attributes
        assert len(attributeset) == len(all_attributes_after_add_2)


    def test_get_attributes(self, client):
        """
            def get_attributes(**kwargs):
        """
        test_attr = JSONObject({
            "name": 'Test Attribute 1',
            "dimension_id": None
        })
        new_attr = client.add_attribute(test_attr)
        attributes = client.get_attributes()
        assert len(attributes) > 0, "get_attributes didn't work as expected!"


    def test_get_template_attributes(self, client):
        """
            SKELETON
            def get_template_attributes(template_id, **kwargs):
        """
        pass

    def test_get_attribute_by_id(self, client, attribute):

        existing_attr = attribute

        retrieved_attr = client.get_attribute_by_id(existing_attr.id)

        assert existing_attr.name           == retrieved_attr.name
        assert existing_attr.dimension_id   == retrieved_attr.dimension_id
        assert existing_attr.description    == retrieved_attr.description

    def test_get_attributes_by_id(self, client):

        test_attrs = [
            JSONObject({
                "name": 'Test Attribute 1',
                "dimension_id": None
            }),
            JSONObject({
                "name": 'Test Attribute 2',
                "dimension_id": 1
            })
        ]
        new_attrs_list = client.add_attributes(test_attrs)

        retrieved_attrs = client.get_attributes_by_id([a.id for a in new_attrs_list])

        assert retrieved_attrs[0].name == 'Test Attribute 1'
        assert retrieved_attrs[1].name == 'Test Attribute 2'

        retrieved_attrs = client.get_attributes_by_id([])
        assert len(retrieved_attrs) == 0


    def test_get_all_attributes(self, client, attributes):

        all_attributes = client.get_attributes()
        attribute_names = []
        for a in all_attributes:
            attribute_names.append(a.name)

        assert "Multi-added Attr 1" in attribute_names
        assert "Multi-added Attr 2" in attribute_names


    def test_get_attribute_by_name_and_dimension(self, client, attribute):
        existing_attr = attribute
        retrieved_attr = client.get_attribute_by_name_and_dimension(
                                            existing_attr.name,
                                            existing_attr.dimension_id)

        assert existing_attr.id == retrieved_attr.id
        assert existing_attr.description == retrieved_attr.description

    def test_check_attr_dimension(self, client, new_dataset):
        """
            def check_attr_dimension(attr_id, **kwargs):
        """
        test_attr = JSONObject({
            "name": 'Test Attribute 1',
            "dimension_id": client.get_dimension_by_unit_id(new_dataset.unit_id).id
        })
        new_attr = client.add_attribute(test_attr)

        result = client.check_attr_dimension(new_attr.id)
        log.info(result)
        assert result == 'OK', "check_attr_dimension didn't work as expected"


        pass

class TestResourceAttribute:

    def test_add_resource_attributes(self,
                                     client,
                                     network_with_data,
                                     attribute):

        new_attr = attribute

        existing_attr = network_with_data.attributes[0]

        #add one new one, plus one existing one. This should result in only one being added
        newattributes = [
            {"attr_id": new_attr.id, "network_id": network_with_data.id, "attr_is_var": "Y"},
            existing_attr
        ]

        added_attrs = client.add_resource_attributes(newattributes)

        updated_network = client.get_network(network_with_data.id)

        assert (network_with_data.id, new_attr.id) in added_attrs
        assert len(updated_network.attributes) == len(network_with_data.attributes) + len(added_attrs)

        assert new_attr.id in [netattr.attr_id for netattr in updated_network.attributes]

    def test_update_resource_attribute(self, client):
        """
            SKELETON
            def update_resource_attribute(resource_attr_id, is_var, **kwargs):
        """
        pass

    def test_delete_resource_attribute(self, client):
        """
            SKELETON
            def delete_resource_attribute(resource_attr_id, **kwargs):
        """
        pass


    def test_add_resource_attrs_from_type(self, client):
        """
            SKELETON
            def add_resource_attrs_from_type(type_id, resource_type, resource_id,**kwargs):
        """
        pass

    def test_get_resource_attribute(self, client):
        """
            SKELETON
            def get_resource_attribute(resource_attr_id, **kwargs):
        """
        pass

    def test_get_all_resource_attributes(self, client):
        """
            SKELETON
            def get_all_resource_attributes(ref_key, network_id, template_id=None, **kwargs):
        """
        pass

    def test_get_resource_attributes(self, client):
        """
            SKELETON
            def get_resource_attributes(ref_key, ref_id, type_id=None, **kwargs):
        """
        pass

    def test_get_all_network_attributes(self, client, network_with_data):
        all_network_attributes = client.get_all_network_attributes(network_with_data.id)

        manual_all_network_attributes = [a.attr_id for a in network_with_data.attributes]
        for n in network_with_data.nodes:
            for a in n.attributes:
                if a.attr_id not in manual_all_network_attributes:
                    manual_all_network_attributes.append(a.attr_id)
        for n in network_with_data.links:
            for a in n.attributes:
                print(a.attr_id)
                if a.attr_id not in manual_all_network_attributes:
                    manual_all_network_attributes.append(a.attr_id)
        for n in network_with_data.resourcegroups:
            for a in n.attributes:
                if a.attr_id not in manual_all_network_attributes:
                    manual_all_network_attributes.append(a.attr_id)

        assert len(set([a.id for a in all_network_attributes])) == len(manual_all_network_attributes)


    def test_add_group_attribute(self, client, network_with_data, attribute):
        group = network_with_data.resourcegroups[0]
        client.add_resource_attribute('GROUP', group.id, attribute.id, 'Y')
        group_attrs = client.get_resource_attributes('GROUP', group.id)
        group_attr_ids = []
        for ga in group_attrs:
            group_attr_ids.append(ga.attr_id)
        assert attribute.id in group_attr_ids

    def test_get_all_group_attributes(self, client, network_with_data):

        #Get all the node attributes in the network
        group_attr_ids = []
        for g in network_with_data.resourcegroups:
            for ga in g.attributes:
                group_attr_ids.append(ga.id)

        group_attributes = client.get_all_resource_attributes('GROUP', network_with_data.id)

        #Check that the retrieved attributes are in the list of group attributes
        retrieved_ras = []
        for ga in group_attributes:
            retrieved_ras.append(ga.id)
        assert set(group_attr_ids) == set(retrieved_ras)



    def test_add_link_attribute(self, client, network_with_data, attribute):
        link = network_with_data.links[0]
        client.add_resource_attribute('LINK', link.id, attribute.id, 'Y')
        link_attributes = client.get_resource_attributes('LINK', link.id)
        network_attr_ids = []

        for ra in link_attributes:
            network_attr_ids.append(ra.attr_id)
        assert attribute.id in network_attr_ids

    def test_get_all_link_attributes(self, client, network_with_data):

        #Get all the node attributes in the network
        link_attr_ids = []
        for l in network_with_data.links:
            for la in l.attributes:
                link_attr_ids.append(la.id)
        link_attributes = client.get_all_resource_attributes('LINK', network_with_data.id)
        #Check that the retrieved attributes are in the list of node attributes
        retrieved_ras = []
        for la in link_attributes:
            retrieved_ras.append(la.id)
        assert set(link_attr_ids) == set(retrieved_ras)


    def test_add_node_attribute(self, client, network_with_data, attribute):
        node = network_with_data.nodes[0]
        client.add_resource_attribute('NODE', node.id, attribute.id, 'Y')
        node_attributes = client.get_resource_attributes('NODE', node.id)
        network_attr_ids = []
        for ra in node_attributes:
            network_attr_ids.append(ra.attr_id)
        assert attribute.id in network_attr_ids

    def test_add_duplicate_node_attribute(self, client, network_with_data, attribute):
        node = network_with_data.nodes[0]
        client.add_resource_attribute('NODE', node.id, attribute.id, 'Y')
        node_attributes = client.get_resource_attributes('NODE', node.id)
        node_attr_ids = []
        for ra in node_attributes:
            node_attr_ids.append(ra.attr_id)
        assert attribute.id in node_attr_ids

        with pytest.raises(hb.HydraError):
            client.add_resource_attribute('NODE', node.id, attribute.id, 'Y')

    def test_get_all_node_attributes(self, client, network_with_data):

        #Get all the node attributes in the network
        node_attr_ids = []
        for n in network_with_data.nodes:
            for a in n.attributes:
                node_attr_ids.append(a.id)

        node_attributes = client.get_all_resource_attributes('NODE', network_with_data.id)

        #Check that the retrieved attributes are in the list of node attributes
        retrieved_ras = []
        for na in node_attributes:
            retrieved_ras.append(na.id)

        assert set(node_attr_ids) == set(retrieved_ras)

        template_id = network_with_data.types[0].template_id

        node_attributes = client.get_all_resource_attributes('NODE', network_with_data.id, template_id)

        #Check that the retrieved attributes are in the list of node attributes
        retrieved_ras = []
        for na in node_attributes:
            retrieved_ras.append(na.id)
        assert set(node_attr_ids) == set(retrieved_ras)



    def test_add_network_attribute(self, client, network_with_data, attribute):

        new_attr = attribute

        client.add_resource_attribute('NETWORK', network_with_data.id, new_attr.id, 'Y')

        updated_network = client.get_network(network_with_data.id)

        network_attr_ids = []

        for ra in updated_network.attributes:
            network_attr_ids.append(ra.attr_id)
        assert new_attr.id in network_attr_ids

    def test_add_network_attrs_from_type(self, client, network_with_data, attribute):
        """
            Recreate the situation where a template is updated, so the network needs.
            to be updated to reflect this change.
            Equivalent to 'apply_template_to_network', just performed differently.
        """

        network = network_with_data

        type_id = network.types[0].id

        #Get the template type, and add a new typeattr to it
        templatetype_j = JSONObject(client.get_templatetype(type_id))

        new_typeattr = JSONObject({
            'attr_id' : attribute.id
        })

        templatetype_j.typeattrs.append(new_typeattr)

        client.update_templatetype(templatetype_j)

        #Record the network's resource attributes before the update
        before_net_attrs = []
        for ra in network.attributes:
            before_net_attrs.append(ra.attr_id)
            log.info("old: %s",ra.attr_id)

        #Set any new resource attributes
        client.add_resource_attrs_from_type(type_id, 'NETWORK', network.id)

        updated_network = client.get_network(network.id)
        after_net_attrs = []
        for ra in updated_network.attributes:
            after_net_attrs.append(ra.attr_id)
            log.info("new: %s",ra.attr_id)

        assert len(after_net_attrs) == len(before_net_attrs) + 1

    def OLD_test_get_network_attrs(self, client, network_with_data):

        net_attrs = client.get_resource_attributes('NETWORK', network_with_data.id)
        net_type_attrs = client.get_resource_attributes('NETWORK',
                                                        network_with_data.id,
                                                        network_with_data.types[0].id)

        assert len(net_attrs) == 3
        assert len(net_type_attrs) == 2

    def test_delete_all_duplicate_attributes(self, client, network_with_data):

        duplicate_attribute = JSONObject({'name': 'duplicate', 'dimension_id': None})

        #use dedicated testing function  which allows duplicates
        dupe_attr_1 = client.add_attribute(duplicate_attribute, check_existing=False)
        dupe_attr_2 = client.add_attribute(duplicate_attribute, check_existing=False)

        all_attrs = client.get_attributes()

        assert dupe_attr_1.id in [a.id for a in all_attrs]
        assert dupe_attr_2.id in [a.id for a in all_attrs]

        #add duplicate resource attributes
        client.add_resource_attribute('NETWORK', network_with_data.id, dupe_attr_1.id, 'Y')
        client.add_resource_attribute('NETWORK', network_with_data.id, dupe_attr_2.id, 'Y')

        #check the dupes are there
        updated_net = client.get_network(network_with_data.id)
        updated_net_ras = [ra.attr_id for ra in updated_net.attributes]
        assert dupe_attr_1.id in updated_net_ras
        assert dupe_attr_2.id in updated_net_ras

        #now add duplicate attrs to the template type
        templatetype_to_update = network_with_data.types[0].id
        client.add_typeattr(JSONObject({'type_id': templatetype_to_update,
                                        'attr_id': dupe_attr_1.id}))
        client.add_typeattr(JSONObject({'type_id': templatetype_to_update,
                                        'attr_id': dupe_attr_2.id}))

        #check the dupes are there
        updated_type = client.get_templatetype(templatetype_to_update)
        assert dupe_attr_1.id in [ta.attr_id for ta in updated_type.typeattrs]
        assert dupe_attr_1.id in [ta.attr_id for ta in updated_type.typeattrs]

        client.delete_all_duplicate_attributes()

        #check the dupes are gone
        updated_net = client.get_network(network_with_data.id)
        updated_net_ras = [ra.attr_id for ra in updated_net.attributes]
        assert dupe_attr_1.id in updated_net_ras
        assert dupe_attr_2.id not in updated_net_ras

        #check the dupes are gone
        updated_type = client.get_templatetype(templatetype_to_update)
        assert dupe_attr_1.id in [ta.attr_id for ta in updated_type.typeattrs]
        assert dupe_attr_2.id not in [ta.attr_id for ta in updated_type.typeattrs]

        reduced_attrs = client.get_attributes()

        #check that the first attr is there, but the dupe is not.
        #the one to keep should be the one with the lowest ID
        lowest_id = min(dupe_attr_1.id, dupe_attr_2.id)
        assert lowest_id in [a.id for a in reduced_attrs]
        assert dupe_attr_2.id not in [a.id for a in reduced_attrs]

    def test_delete_duplicate_resourceattributes(self, client, network_with_data):

        #first add a duplicate resourceattr to the network

        #find a resourceattr.
        network_with_data.attributes.sort(key=lambda x : x.attr_id)
        ra1 = network_with_data.attributes[0]
        ra2 = network_with_data.attributes[1]
        ra3 = network_with_data.attributes[-1] # not associated to a template

        ra_attribute1 = client.get_attribute_by_id(ra1.attr_id)
        ra_attribute2 = client.get_attribute_by_id(ra2.attr_id)
        ra_attribute3 = client.get_attribute_by_id(ra3.attr_id)

        #create an attribute with the same name but a different dimension
        duplicate_attribute1 = JSONObject({
            'name': ra_attribute1.name,
            'dimension_id': 1
        })
        duplicate_attribute2 = JSONObject({
            'name': ra_attribute2.name,
            'dimension_id': 1
        })

        duplicate_attribute3 = JSONObject({
            'name': ra_attribute3.name,
            'dimension_id': 1
        })
        dupeattr1 = client.add_attribute(duplicate_attribute1)
        dupeattr2 = client.add_attribute(duplicate_attribute2)
        dupeattr3 = client.add_attribute(duplicate_attribute3)

        dupe_ra1 = client.add_resource_attribute('NETWORK', network_with_data.id, dupeattr1.id, 'N')#
        #get an arbitrary dataset
        dataset = client.get_dataset(1)
        #set a value on the RA which sould get transferred in the deletion later
        new_rscen = client.add_data_to_attribute(network_with_data.scenarios[0].id,
                                                 dupe_ra1.id,
                                                 dataset)
        #add 2 more dupes but with no data associated to them
        dupe_ra2 = client.add_resource_attribute('NETWORK', network_with_data.id, dupeattr2.id, 'N')
        dupe_ra3 = client.add_resource_attribute('NETWORK', network_with_data.id, dupeattr3.id, 'N')

        updated_net = client.get_network(network_with_data.id, include_data=True)
        updated_net_ras = [ra.attr_id for ra in updated_net.attributes]
        assert dupeattr1.id in updated_net_ras
        assert dupeattr2.id in updated_net_ras
        assert dupeattr3.id in updated_net_ras

        #verify the data has been associated to the dupe RA
        original_rs = network_with_data.scenarios[0].resourcescenarios
        new_rs = updated_net.scenarios[0].resourcescenarios
        assert len(new_rs) == len(original_rs) + 1
        assert ra1.id not in [rs.resource_attr_id for rs in new_rs]
        assert dupe_ra1.id in [rs.resource_attr_id for rs in new_rs]

        #now that the new attribute is added, try to delete it.
        client.delete_duplicate_resourceattributes()

        updated_net = client.get_network(network_with_data.id, include_data=True)
        updated_net_ras = [ra.attr_id for ra in updated_net.attributes]
        #the number of network attributes has decreased because BOTH duplicates
        #of RA3 (which are not associated to a template) have been removed.
        #This means one of the original ones is now gone
        assert len(updated_net.attributes) == len(network_with_data.attributes) -1
        assert dupeattr1.id not in updated_net_ras
        assert dupeattr2.id not in updated_net_ras
        assert dupeattr3.id not in updated_net_ras

        #verify the data which was on the dupe has been remapped to the remaining, correct, attribute
        original_rs = network_with_data.scenarios[0].resourcescenarios
        new_rs = updated_net.scenarios[0].resourcescenarios
        assert len(new_rs) == len(original_rs) + 1
        assert ra1.id in [rs.resource_attr_id for rs in new_rs]
        assert dupe_ra1.id not in [rs.resource_attr_id for rs in new_rs]

class TestAttributeMap:

    def test_set_attribute_mapping(self, client, networkmaker):
        net1 = networkmaker.create()
        net2 = networkmaker.create()
        net3 = networkmaker.create()

        s1 = net1.scenarios[0]
        s2 = net2.scenarios[0]

        node_1 = net1.nodes[0]
        node_2 = net2.nodes[1]
        node_3 = net3.nodes[2]

        attr_1 = client.testutils.get_by_name('node_attr_a', node_1.attributes)
        attr_2 = client.testutils.get_by_name('node_attr_b', node_2.attributes)
        attr_3 = client.testutils.get_by_name('node_attr_c', node_3.attributes)

        rs_to_update_from = None
        for rs in s1.resourcescenarios:
            if rs.resource_attr_id == attr_1.id:
                rs_to_update_from = rs


        rs_to_change = None
        for rs in s2.resourcescenarios:
            if rs.resource_attr_id == attr_2.id:
                rs_to_change = rs

        client.set_attribute_mapping(attr_1.id, attr_2.id)
        client.set_attribute_mapping(attr_1.id, attr_3.id)


        all_mappings_1 = client.get_mappings_in_network(net1.id)
        all_mappings_2 = client.get_mappings_in_network(net2.id, net2.id)


        assert len(all_mappings_1) == 2
        assert len(all_mappings_2) == 1

        node_mappings_1 = client.get_node_mappings(node_1.id)
        node_mappings_2 = client.get_node_mappings(node_1.id, node_2.id)

        assert len(node_mappings_1) == 2
        assert len(node_mappings_2) == 1

        map_exists = client.check_attribute_mapping_exists(attr_1.id, attr_2.id)
        assert map_exists == 'Y'
        map_exists = client.check_attribute_mapping_exists(attr_2.id, attr_1.id)
        assert map_exists == 'N'
        map_exists = client.check_attribute_mapping_exists(attr_2.id, attr_3.id)
        assert map_exists == 'N'


        updated_rs = client.update_value_from_mapping(attr_1.id, attr_2.id, s1.id, s2.id)

        assert str(updated_rs.dataset.value) == str(rs_to_update_from.dataset.value)

        log.info("Deleting %s -> %s", attr_1.id, attr_2.id)
        client.delete_attribute_mapping(attr_1.id, attr_2.id)
        all_mappings_1 = client.get_mappings_in_network(net1.id)
        assert len(all_mappings_1) == 1

        client.delete_mappings_in_network(net1.id)
        all_mappings_1 = client.get_mappings_in_network(net1.id)
        assert len(all_mappings_1) == 0



    def test_delete_attribute_mapping(self, client):
        """
            SKELETON
            def delete_attribute_mapping(resource_attr_a, resource_attr_b, **kwargs):
        """
        pass
    def test_delete_mappings_in_network(self, client):
        """
            SKELETON
            def delete_mappings_in_network(network_id, network_2_id=None, **kwargs):
        """
        pass
    def test_get_mappings_in_network(self, client):
        """
            SKELETON
            def get_mappings_in_network(network_id, network_2_id=None, **kwargs):
        """
        pass
    def test_get_node_mappings(self, client):
        """
            SKELETON
            def get_node_mappings(node_id, node_2_id=None, **kwargs):
        """
        pass
    def test_get_link_mappings(self, client):
        """
            SKELETON
            def get_link_mappings(link_id, link_2_id=None, **kwargs):
        """
        pass
    def test_get_network_mappings(self, client):
        """
            SKELETON
            def get_network_mappings(network_id, network_2_id=None, **kwargs):
        """
        pass
    def test_check_attribute_mapping_exists(self, client):
        """
            SKELETON
            def check_attribute_mapping_exists(resource_attr_id_source, resource_attr_id_target, **kwargs):
        """
        pass











class TestAttributeGroups:
    """
        Test for attribute Groups-based functionality
    """


    def test_add_attribute_group(self, client, projectmaker, attribute):
        project = projectmaker.create()

        newgroup = JSONObject({
            'project_id'  : project.id,
            'name'        : "Attribute Group %s" % (datetime.datetime.now(),),
            'description' : "A description of an attribute group",
            'layout'      : {"color": "green"},
            'exclusive'   : 'Y',
        })

        newgroup = client.add_attribute_group(newgroup)

        retrieved_new_group = client.get_attribute_group(newgroup.id)

        assert retrieved_new_group.name == newgroup.name

    def test_update_attribute_group(self, client, attributegroup):

        newname = attributegroup.name + " Updated"

        attributegroup.name = newname

        client.update_attribute_group(attributegroup)

        retrieved_new_group = client.get_attribute_group(attributegroup.id)

        assert retrieved_new_group.name == newname


    def test_delete_attribute_group(self, client, attributegroup):

        client.delete_attribute_group(attributegroup.id)

        with pytest.raises(hb.HydraError):
            client.get_attribute_group(attributegroup.id)

    def test_basic_add_attribute_group_items(self, client, projectmaker, network_with_data, attributegroupmaker):
        project = projectmaker.create()

        #convenience naming
        network = network_with_data

        #Create two groups -- attributes which are associated with a network,
        #and everything else.
        group_1 = attributegroupmaker.create(project.id, "Network Attributes")
        group_2 = attributegroupmaker.create(project.id, "Node Attributes")

        network_attributes = []
        for netattr in network.attributes:
            network_attributes.append(JSONObject(
                {'attr_id' : netattr.attr_id,
                 'network_id' : network.id,
                 'group_id' : group_1.id
                 }))

        node_attr_tracker = []
        node_attributes = []
        for node in network.nodes:
            for node_attr in node.attributes:
                if node_attr.attr_id not in node_attr_tracker:
                    node_attributes.append(JSONObject(
                        {'attr_id' : node_attr.attr_id,
                         'network_id' : network.id,
                         'group_id' : group_2.id
                         }))
                    node_attr_tracker.append(node_attr.attr_id)


        client.add_attribute_group_items(network_attributes)

        client.add_attribute_group_items(node_attributes)

        all_items_in_network = client.get_network_attributegroup_items(network.id)


        assert len(all_items_in_network) == len(network_attributes)+len(node_attributes)

    def test_exclusive_add_attribute_group_items(self, client, projectmaker, network_with_data, attributegroupmaker):
        """
            add attributes to a group that are already in an exclusive group
        """

        project = projectmaker.create()

        #convenience naming
        network = network_with_data

        #Create two groups -- attributes which are associated with a network,
        #and everything else.
        group_1 = attributegroupmaker.create(project.id, "Network Attributes", 'Y')
        group_2 = attributegroupmaker.create(project.id, "Node Attributes")

        network_attributes = []
        node_attributes = []

        for netattr in network.attributes:
            network_attributes.append(JSONObject(
                {'attr_id' : netattr.attr_id,
                 'network_id' : network.id,
                 'group_id' : group_1.id}))
            #Put these attributes into both groups. THis should fail, as group 1
            #is exclusive, and already has these attributes
            node_attributes.append(JSONObject(
                {'attr_id' : netattr.attr_id,
                 'network_id' : network.id,
                 'group_id' : group_2.id}))


        node_attr_tracker = []
        for node in network.nodes:
            for node_attr in node.attributes:
                if node_attr.attr_id not in node_attr_tracker:
                    node_attributes.append(JSONObject(
                        {'attr_id' : node_attr.attr_id,
                         'network_id' : network.id,
                         'group_id' : group_2.id}))
                    node_attr_tracker.append(node_attr.attr_id)


        log.info("Adding items to group 1 (network attributes)")
        client.add_attribute_group_items(network_attributes)

        #add a group with attributes that are already in an exclusive group
        with pytest.raises(hb.HydraError):
            log.info("Adding items to group 2 (node attributes, plus network attributes)")
            client.add_attribute_group_items(node_attributes)

    def test_reverse_exclusive_add_attribute_group_items(self, client, projectmaker, network_with_data, attributegroupmaker):
        """
            add attributes to an exclusive group that are already in another group
        """

        project = projectmaker.create()

        #convenience naming
        network = network_with_data

        #Create two groups -- attributes which are associated with a network,
        #and everything else.
        group_1 = attributegroupmaker.create(project.id, "Network Attributes", 'Y')
        group_2 = attributegroupmaker.create(project.id, "Node Attributes")

        network_attributes = []
        node_attributes = []

        for netattr in network.attributes:
            network_attributes.append(JSONObject(
                {'attr_id' : netattr.attr_id,
                 'network_id' : network.id,
                 'group_id' : group_1.id}))
            #Put these attributes into both groups. THis should fail, as group 1
            #is exclusive, and already has these attributes
            node_attributes.append(JSONObject(
                {'attr_id' : netattr.attr_id,
                 'network_id' : network.id,
                 'group_id' : group_2.id}))


        node_attr_tracker = []
        for node in network.nodes:
            for node_attr in node.attributes:
                if node_attr.attr_id not in node_attr_tracker:
                    node_attributes.append(JSONObject(
                        {'attr_id' : node_attr.attr_id,
                         'network_id' : network.id,
                         'group_id' : group_2.id}))
                    node_attr_tracker.append(node_attr.attr_id)


        log.info("Adding items to group 2 (node attributes, plus network attributes)")
        client.add_attribute_group_items(node_attributes)

        #add attributes to an exclusive group that are already in another group
        with pytest.raises(hb.HydraError):
            log.info("Adding items to group 1 (network attributes)")
            client.add_attribute_group_items(network_attributes)

    def test_delete_attribute_group_items(self, client, projectmaker, network_with_data, attributegroupmaker):
        project = projectmaker.create()

        #convenience naming
        network = network_with_data

        #Create two groups -- attributes which are associated with a network,
        #and everything else.
        group_1 = attributegroupmaker.create(project.id, "Network Attributes")
        group_2 = attributegroupmaker.create(project.id, "Node Attributes")

        network_attributes = []
        for netattr in network.attributes:
            network_attributes.append(JSONObject(
                {'attr_id' : netattr.attr_id,
                 'network_id' : network.id,
                 'group_id' : group_1.id}))

        node_attr_tracker = []
        node_attributes   = []
        for node in network.nodes:
            for node_attr in node.attributes:
                if node_attr.attr_id not in node_attr_tracker:
                    node_attributes.append(JSONObject(
                        {'attr_id' : node_attr.attr_id,
                         'network_id' : network.id,
                         'group_id' : group_2.id}))
                    node_attr_tracker.append(node_attr.attr_id)


        client.add_attribute_group_items(network_attributes)

        client.add_attribute_group_items(node_attributes)

        all_items_in_network = client.get_network_attributegroup_items(network.id)

        assert len(all_items_in_network) == len(network_attributes)+len(node_attributes)

        #Now remove all the node attributes
        client.delete_attribute_group_items(node_attributes)

        all_items_in_network = client.get_network_attributegroup_items(network.id)

        assert len(all_items_in_network) == len(network_attributes)

    def test_get_attribute_group(self, client):
        """
            SKELETON
            def get_attribute_group(group_id, **kwargs):
        """
        pass
    def test_get_network_attributegroup_items(self, client):
        """
            SKELETON
            def get_network_attributegroup_items(network_id, **kwargs):
        """
        pass
    def test_get_group_attributegroup_items(self, client):
        """
            SKELETON
            def get_group_attributegroup_items(network_id, group_id, **kwargs):
        """
        pass
    def test_get_attribute_item_groups(self, client):
        """
            SKELETON
            def get_attribute_item_groups(network_id, attr_id, **kwargs):
        """
        pass
    def test_add_attribute_group_items(self, client):
        """
            SKELETON
            def add_attribute_group_items(attributegroupitems, **kwargs):
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from ..exceptions import HydraError, ResourceNotFoundError
import logging
log = logging.getLogger(__name__)
from .. import db
from ..db.model import Network, Project, ProjectOwner, NetworkOwner, User, Dataset
from hydra_base.lib.objects import JSONObject
from sqlalchemy.orm.exc import NoResultFound
from hydra_base.util.permissions import required_role

def _get_project(project_id):
    try:
        proj_i = db.DBSession.query(Project).filter(Project.id==project_id).one()
        return proj_i
    except NoResultFound:
        raise ResourceNotFoundError("Project %s not found"%(project_id,))

def _get_network(network_id):
    try:
        net_i = db.DBSession.query(Network).filter(Network.id == network_id).one()
        return net_i
    except NoResultFound:
        raise ResourceNotFoundError("Network %s not found"%(network_id))

def _get_user(username):
    try:
        user_i = db.DBSession.query(User).filter(User.username==username).one()
        return user_i
    except NoResultFound:
        raise ResourceNotFoundError("User %s not found"%(username))

def _get_dataset(dataset_id):
    try:
        dataset_i = db.DBSession.query(Dataset).filter(Dataset.id==dataset_id).one()
        return dataset_i
    except NoResultFound:
        raise ResourceNotFoundError("Dataset %s not found"%(dataset_id))

def share_network(network_id, usernames, read_only, share, **kwargs):
    """
        Share a network with a list of users, identified by their usernames.

        The read_only flag ('Y' or 'N') must be set
        to 'Y' to allow write access or sharing.

        The share flat ('Y' or 'N') must be set to 'Y' to allow the
        project to be shared with other users
    """

    user_id = kwargs.get('user_id')
    net_i = _get_network(network_id)
    net_i.check_share_permission(user_id)

    #Support wither boolean or character based flags for read_only and share
    if read_only in ('Y', True):
        write = 'N'
        share = 'N'
    else:
        write = 'Y'

    if share is True:
        share = 'Y'
    elif share is False:
        share = 'N'

    if net_i.created_by != int(user_id) and share == 'Y':
        raise HydraError(f"Cannot share the 'sharing' ability as user {user_id} is not"
                         f" the owner of network {network_id}")

    for username in usernames:
        user_i = _get_user(username)
        #Set the owner ship on the network itself
        net_i.set_owner(user_i.id, write=write, share=share)

        Project.clear_cache(user_i.id)
    db.DBSession.flush()

def unshare_network(network_id, usernames, **kwargs):
    """
        Un-Share a network with a list of users, identified by their usernames.
    """

    user_id = kwargs.get('user_id')
    net_i = _get_network(network_id)
    net_i.check_share_permission(user_id)

    for username in usernames:
        user_i = _get_user(username)
        #Set the owner ship on the network itself
        net_i.unset_owner(user_i.id)
        Project.clear_cache(user_i.id)

    db.DBSession.flush()

def share_project(project_id, usernames, read_only=False, share=False, flush=True, **kwargs):
    """
        Share an entire project with a list of users, identifed by
        their usernames.

        The read_only flag ('Y' or 'N') must be set
        to 'Y' to allow write access or sharing.

        The share flat ('Y' or 'N') must be set to 'Y' to allow the
        project to be shared with other users
    """
    user_id = kwargs.get('user_id')

    proj_i = _get_project(project_id)

    #Is the sharing user allowed to share this project?
    proj_i.check_share_permission(int(user_id))

    user_id = int(user_id)

    # for owner in proj_i.owners:
    #     if user_id == owner.user_id:
    #         break
    # else:
    #     raise HydraError("Permission Denied. Cannot share project.")

    if read_only in ('Y', True):
        write = 'N'
        share = 'N'
    if read_only in ('N', False):
        write = 'Y'
    if read_only is None:
        write = None

    if share in ('Y', True):
        share = 'Y'
    if share in ('N', False):
        share = 'N'

    # if proj_i.created_by != user_id and share == 'Y':
    #     raise HydraError("Cannot share the 'sharing' ability as user %s is not"
    #                  " the owner of project %s"%
    #                  (user_id, project_id))

    for username in usernames:
        user_i = _get_user(username)

        proj_i.set_owner(user_i.id, write=write, share=share)

        Project.clear_cache(user_i.id)


    if flush is True:
        db.DBSession.flush()

def unshare_project(project_id, usernames, **kwargs):
    """
        Un-share a project with a list of users, identified by their usernames.
    """

    user_id = kwargs.get('user_id')
    proj_i = _get_project(project_id)
    proj_i.check_share_permission(user_id)

    for username in usernames:
        user_i = _get_user(username)
        #Set the owner ship on the network itself
        proj_i.unset_owner(user_i.id)
        Project.clear_cache(user_i.id)
    db.DBSession.flush()

def set_project_permission(project_id, usernames, read, write, share,**kwargs):
    """
        Set permissions on a project to a list of users, identifed by
        their usernames.

        The read flag ('Y' or 'N') sets read access, the write
        flag sets write access. If the read flag is 'N', then there is
        automatically no write access or share access.
    """
    user_id = kwargs.get('user_id')

    proj_i = _get_project(project_id)

    #Is the sharing user allowed to share this project?
    proj_i.check_share_permission(user_id)

    #You cannot edit something you cannot see.
    if read == 'N':
        write = 'N'
        share = 'N'

    for username in usernames:
        user_i = _get_user(username)

        #The creator of a project must always have read and write access
        #to their project
        if proj_i.created_by == user_i.id:
            raise HydraError("Cannot set permissions on project %s"
                             " for user %s as this user is the creator." %
                             (project_id, username))

        proj_i.set_owner(user_i.id, read=read, write=write)

        for net_i in proj_i.networks:
            net_i.set_owner(user_i.id, read=read, write=write, share=share)
    db.DBSession.flush()

def set_network_permission(network_id, usernames, read, write, share,**kwargs):
    """
        Set permissions on a network to a list of users, identifed by
        their usernames. The read flag ('Y' or 'N') sets read access, the write
        flag sets write access. If the read flag is 'N', then there is
        automatically no write access or share access.
    """

    user_id = kwargs.get('user_id')

    net_i = _get_network(network_id)

    #Check if the user is allowed to share this network.
    net_i.check_share_permission(user_id)

    #You cannot edit something you cannot see.
    if read == 'N':
        write = 'N'
        share = 'N'

    for username in usernames:

        user_i = _get_user(username)

        #The creator of a network must always have read and write access
        #to their project
        if net_i.created_by == user_i.id:
            raise HydraError("Cannot set permissions on network %s"
                             " for user %s as tis user is the creator." %
                             (network_id, username))

        net_i.set_owner(user_i.id, read=read, write=write, share=share)
    db.DBSession.flush()

def hide_dataset(dataset_id, exceptions, read, write, share,**kwargs):
    """
        Hide a particular piece of data so it can only be seen by its owner.
        Only an owner can hide (and unhide) data.
        Data with no owner cannot be hidden.

        The exceptions paramater lists the usernames of those with permission to view the data
        read, write and share indicate whether these users can read, edit and share this data.
    """

    user_id = kwargs.get('user_id')
    dataset_i = _get_dataset(dataset_id)

    #check that I can hide the dataset
    if dataset_i.created_by != int(user_id):
        raise HydraError('Permission denied. '
                        'User %s is not the owner of dataset %s'
                        %(user_id, dataset_i.name))

    dataset_i.hidden = 'Y'
    if exceptions is not None:
        for username in exceptions:
            user_i = _get_user(username)
            dataset_i.set_owner(user_i.id, read=read, write=write, share=share)
    db.DBSession.flush()

def unhide_dataset(dataset_id,**kwargs):
    """
        Hide a particular piece of data so it can only be seen by its owner.
        Only an owner can hide (and unhide) data.
        Data with no owner cannot be hidden.

        The exceptions paramater lists the usernames of those with permission to view the data
        read, write and share indicate whether these users can read, edit and share this data.
    """

    user_id = kwargs.get('user_id')
    dataset_i = _get_dataset(dataset_id)
    #check that I can unhide the dataset
    if dataset_i.created_by != int(user_id):
        raise HydraError('Permission denied. '
                        'User %s is not the owner of dataset %s'
                        %(user_id, dataset_i.name))

    dataset_i.hidden = 'N'
    db.DBSession.flush()

@required_role("admin")
def get_all_project_owners(project_ids=None, **kwargs):
    """
        Get the project owner entries for all the requested projects.
        If the project_ids argument is None, return all the owner entries
        for ALL projects
    """


    projowner_qry = db.DBSession.query(ProjectOwner)

    if project_ids is not None:
       projowner_qry = projowner_qry.filter(ProjectOwner.project_id.in_(project_ids))

    project_owners_i = projowner_qry.all()

    return [JSONObject(project_owner_i) for project_owner_i in project_owners_i]

@required_role("admin")
def get_all_network_owners(network_ids=None, **kwargs):
    """
        Get the network owner entries for all the requested networks.
        If the network_ids argument is None, return all the owner entries
        for ALL networks
    """


    networkowner_qry = db.DBSession.query(NetworkOwner)

    if network_ids is not None:
       networkowner_qry = networkowner_qry.filter(NetworkOwner.network_id.in_(network_ids))

    network_owners_i = networkowner_qry.all()

    return [JSONObject(network_owner_i) for network_owner_i in network_owners_i]

@required_role("admin")
def bulk_set_project_owners(project_owners, **kwargs):
    """
        Set the project owner of multiple projects at once.
        Accepts a list of JSONObjects which look like:
            {
             'project_id': XX,
             'user_id'   : YY,
             'view'      : 'Y'/ 'N'
             'edit'      : 'Y'/ 'N'
             'share'      : 'Y'/ 'N'
            }
           """

    project_ids = [po.project_id for po in project_owners]

    existing_projowners = db.DBSession.query(ProjectOwner).filter(ProjectOwner.project_id.in_(project_ids)).all()

    #Create a lookup based on the unique key for this table (project_id, user_id)
    po_lookup = {}

    for po in existing_projowners:
        po_lookup[(po.project_id, po.user_id)] = po

    for project_owner in project_owners:
        #check if the entry is already there
        if po_lookup.get((project_owner.project_id, project_owner.user_id)):
            continue

        new_po = ProjectOwner()
        new_po.project_id = project_owner.project_id
        new_po.user_id    = project_owner.user_id
        new_po.view       = project_owner.view
        new_po.edit       = project_owner.edit
        new_po.share      = project_owner.share

        db.DBSession.add(new_po)

    db.DBSession.flush()

@required_role("admin")
def bulk_set_network_owners(network_owners, **kwargs):
    """
        Set the network owner of multiple networks at once.
        Accepts a list of JSONObjects which look like:
            {
             'network_id': XX,
             'user_id'   : YY,
             'view'      : 'Y'/ 'N'
             'edit'      : 'Y'/ 'N'
             'share'      : 'Y'/ 'N'
            }
           """

    network_ids = [no.network_id for no in network_owners]

    existing_projowners = db.DBSession.query(NetworkOwner).filter(NetworkOwner.network_id.in_(network_ids)).all()

    #Create a lookup based on the unique key for this table (network_id, user_id)
    no_lookup = {}

    for no in existing_projowners:
        no_lookup[(no.network_id, no.user_id)] = no

    for network_owner in network_owners:
        #check if the entry is already there
        if no_lookup.get((network_owner.network_id, network_owner.user_id)):
            continue

        new_no = NetworkOwner()
        new_no.network_id = network_owner.network_id
        new_no.user_id    = network_owner.user_id
        new_no.view       = network_owner.view
        new_no.edit       = network_owner.edit
        new_no.share      = network_owner.share

        db.DBSession.add(new_no)

    db.DBSession.flush()



"""
This is a utility which takes a source network ID and target network ID,
and sets the types from the source onto the target -- this is needed when
the resource types for the target network don't exist.

The names of the nodes in both networks need to be the same
"""

import click
import hydra_base as hb
from hydra_base.lib.objects import JSONObject
from hydra_base.db.model import *

class TypeRestorer:
    def __init__(self, source_network_id, type_csv, template_id, target_network_id, default_link, debug=False):
        """
            The default link is the default link to use if a link exists in the
            target network but not in the source

            We do not provide a default node argument, because the basic functionality
            relies on the nodes in the target being a subset of the source, and so no
            default should be necessary
        """
        self.source_network_id = source_network_id
        self.target_network_id = target_network_id
        self.type_csv = type_csv
        self.template_id = template_id

        self.check_inputs()

        self.source_network = None
        self.source_nodes = []
        self.source_links = []
        self.source_groups = []

        self.target_network = None
        self.target_nodes = []
        self.target_links = []
        self.target_groups = []

        self.network_type_map = {}
        self.node_type_map = {}
        self.link_type_map = {}
        self.group_type_map = {}

        self.template = None
        self.network_type = None

        self.new_resource_types = {'NODE': 0, 'LINK': 0, 'GROUP': 0, 'NETWORK': 0}

        self.default_link_name = default_link
        self.default_link_type = None

        self.debug = debug # don't commit the transaction if true

    def check_inputs(self):
        """
            make sure the inputs are all valid
        """

        if self.source_network_id is None and self.type_csv is None:
            raise Exception("Please specifiy a '--source_network_id' or a '--type_csv'")

        if self.type_csv is not None and self.template_id is None:
            raise Exception("Cannot use --type-csv without a --template-id argument."
                            " Please specify a --template-id")

        if self.target_network_id is None:
            raise Exception("Please specify a target network ID with '--target_network_id'")

        print("Inputs are valid")

    def load_data(self):

        hb.db.connect()

        self.target_network = hb.db.DBSession.query(Network).filter(
            Network.id == self.target_network_id).one()
        self.target_nodes = hb.db.DBSession.query(Node).filter(
            Node.network_id == self.target_network_id).all()
        self.target_links = hb.db.DBSession.query(Link).filter(
            Link.network_id == self.target_network_id).all()
        self.target_groups = hb.db.DBSession.query(ResourceGroup).filter(
            ResourceGroup.network_id == self.target_network_id).all()

        self.load_type_map()
        self.set_default_link_type()


    def set_default_link_type(self):
        """
            Identify and set the default link tyupe, if specified
        """
        if self.default_link_name is not None:
            for t in self.template.templatetypes:
                if t.name.lower() == self.default_link_name.lower():
                    self.default_link_type = JSONObject({'id': t.id})

            if self.default_link_type is None:
                raise Exception(f"Default Link Type {self.default_link_name} "
                                f"not found in template {self.template.name}")

    def load_type_map(self):
        """
            Create the mapping from node to type
        """

        if self.type_csv is not None:
            self.load_type_map_from_csv()
        else:
            self.load_type_map_from_source_network()

    def load_type_map_from_csv(self):
        """
            Create a mapping from node name to type by getting the template
            and using a type_csv for the mapping
        """

        self.template = hb.db.DBSession.query(Template).filter(
            Template.id==self.template_id).first()

        if self.template is None:
            raise Exception(f"Template {self.template_id} not found")

        #map from type name to type object
        type_name_map = {}
        for tt in self.template.templatetypes:
            type_name_map[tt.name.lower()] = tt

        self.csv_df = pd.read_csv(self.type_csv, index_col='index')

        for nodename, typename in self.csv_df.itertuples():
            typename = typename.lower()
            if typename not in type_name_map:
                raise Exception(f"Type name {typename} not found in template {self.template_id}"
                                f"allowed typenames are: {type_name_map.keys()}")

            self.node_type_map[nodename] = type_name_map[typename]

    def get_network_type(self):
        """
            Look in the template for the template type for the NETWORK
        """
        if self.template_id is None:
            raise Exception("Can't get network type. Template is null")

        for tt in self.template.templatetypes:
            if tt.resource_type == 'NETWORK':
                return tt

        raise Exception(f"Unable to find NETWORK template type in template {self.template.id}")


    def load_type_map_from_source_network(self):
        """
            Create a mapping from node name to type by using a source network
        """
        self.source_network = hb.db.DBSession.query(Network).filter(
            Network.id == self.source_network_id).one()
        self.source_nodes = hb.db.DBSession.query(Node).filter(
            Node.network_id == self.source_network_id).all()
        self.source_links = hb.db.DBSession.query(Link).filter(
            Link.network_id == self.source_network_id).all()
        self.source_groups = hb.db.DBSession.query(ResourceGroup).filter(
            ResourceGroup.network_id == self.source_network_id).all()

        self.template = self.source_network.types[0].templatetype.template

        # SOURCE NODES
        for n in self.source_nodes:
            self.node_type_map[n.name] = n.types[0].templatetype
        # SOURCE LINKS
        for l in self.source_links:
            start_node = l.node_a.name
            end_node   = l.node_b.name
            self.link_type_map[(start_node, end_node)] = l.types[0].templatetype
        # SOURCE GROUPS
        for g in self.source_groups:
            self.group_type_map[g.name] = g.types[0].templatetype

    # generic function to add a resourcetype
    def add_resource_type(self, ref_key, target_resource, source_type):
        """
            Set the resource type of a resource. ref_key helps identify the type
            of resource, target_resource is the resource itself, and source type
            is the templatetype object to use
        """
        for t in target_resource.types:
            if t.type_id == source_type.id:
                print(f"{ref_key} {target_resource.name} already has the correct type")
                break
        else:
            rt = ResourceType()
            rt.ref_key = ref_key
            if ref_key == "GROUP":
                rt.group_id = target_resource.id
            elif ref_key == "NODE":
                rt.node_id = target_resource.id
            elif ref_key == "NETWORK":
                rt.network_id = target_resource.id
            elif ref_key == "LINK":
                rt.link_id = target_resource.id
            else:
                raise Exception(f"The ref_key {ref_key} is not valid!")

            rt.type_id = source_type.id
            rt.child_template_id = source_type.template_id
            self.new_resource_types[ref_key] = self.new_resource_types[ref_key] + 1
            hb.db.DBSession.add(rt)

    def check_networks_compatible(self):
        """
            Check the 2 networks are compatible by testing whether the source network
            contains all the nodes needed by the target network
        """
        #first check we've got compatible networks
        incompatible_nodes = []
        for n in self.target_nodes:
            if n.name not in self.node_type_map:
                incompatible_nodes.append(n.name)
        if len(incompatible_nodes) > 0:
            print(f"***ERROR*** Unable to map types. Found a node name(s) in"
                  f" target network {self.target_network_id} which is "
                  f"not in the source network {self.source_network_id} or type_csv.\n"
                  f"These are:  {incompatible_nodes}")
            return False

        #If a default link type is provided, it means that there are probably links the targtet
        #network not present in the source. If this is the case, then don't bother checking
        #for compatible links
        if self.default_link_type is None:
            for l in self.target_links:
                start_node = l.node_a.name
                end_node   = l.node_b.name
                if (start_node, end_node) not in self.link_type_map:
                    print(f"**ERROR** Unable to map types. Found a link name '{l.name}' in"
                          f" target network {self.target_network_id} which is "
                          f"not in the source network {self.source_network_id}")
                    return False

        for g in self.target_groups:
            if g.name not in self.group_type_map:
                print(f"Unable to map types. Found a group name '{g.name}' in"
                      f" target network {self.target_network_id} which is "
                      f"not in the source network {self.source_network_id}")
                return False
        return True
    def run(self):
        networks_compatible = self.check_networks_compatible()
        if (networks_compatible is False):
            return

        #OK we have compatible networks. Now create some resource types
        self.add_resource_type("NETWORK",
                                   self.target_network,
                                   self.get_network_type())

        for n in self.target_nodes:
            matching_type = self.node_type_map[n.name]
            self.add_resource_type("NODE", n, matching_type)


        for l in self.target_links:
            start_node = l.node_a.name
            end_node   = l.node_b.name
            matching_type = self.link_type_map.get((start_node, end_node), self.default_link_type)
            self.add_resource_type("LINK", l, matching_type)

        for g in self.target_groups:
            matching_type = self.group_type_map[g.name]
            self.add_resource_type("GROUP", g, matching_type)

        try:

            # To reenable later
            if self.debug is not True:
                hb.db.DBSession.flush()
                hb.commit_transaction()
                print("Inserted types: %s"%self.new_resource_types)
            else:
                print("**DEBUG MODE ENABLED** Inserted types: %s"%self.new_resource_types)
                print("**DEBUG MODE ENABLED** Transaction not commited to DB")
                hb.rollback_transaction()

        except Exception as e:
            print("An error has occurred: %s", e)
            hb.rollback_transaction()

@click.command()
@click.option('--source_network_id', type=int, default=None,
              help="(optional, see type_csv) The ID of the working network"+
              " i.e. the one where the nodes have correct types."+
              " This network must have the same node names as the broken network.")
@click.option('--type-csv', type=click.Path(), default=None,
              help="(optional, see source-network-id) A CSV file where the first column is the node names,"+
              "and the second column is the type names of that node."+
              "The columns must be headed 'index' and 'type'"+
              "*This must be used in conjunction with the --template-id argument")
@click.option('--template-id', type=int, default=None,
              help="(optional, see type-csv). This must be included if using the --type-csv"+
              "argument. This is the ID of the template to be applied to the target network")
@click.option('-t', '--target_network_id', type=int, default=None,
              help="The ID of the broken network i.e. the one where the nodes do not have types")
@click.option('--default-link', type=str, default=None,
              help="Name of a default link type, if a link exists in the target but not the source")
@click.option('--debug', is_flag=True, default=False, help="Don't commit the transaction if set.")
def restore(source_network_id=None, type_csv=None, template_id=None, target_network_id=None, default_link=None, debug=False):
    restorer = TypeRestorer(source_network_id, #either this OR
                            type_csv, #this and...
                            template_id,#...this
                            target_network_id, # mandatory
                            default_link=default_link,
                            debug=debug)
    restorer.load_data()
    restorer.run()

if __name__ == '__main__':
    restore()
